# -*- coding: utf-8 -*-

#   This file is part of emesene.
#
#    Emesene is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    emesene is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with emesene; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import gtk
import pango
import emesenelib

class ConversationWindow(gtk.Window):
    '''This class is the conversation window, is contains a textbuffer for the
    input a toolbox and a set of tabs, if there is just one conversation
    the window is like the one in earlier versions but if we add more
    conversations it behaves like a Gaim window'''

    def __init__(self, controller, conversation):
        '''Constructor, the conversation parameter is an instance of Conversation,'''
        
        gtk.Window.__init__(self)
        self.controller = controller
        self.config = controller.config
        self.closed = False
        self.notOpen = True
        self.last_mark = None

        self.width = self.config.user['convWindowWidth']
        self.height = self.config.user['convWindowHeight']
        self.set_default_size(self.width, self.height)
        self.connect('size-allocate', self.on_size_alloc)
        
        # accelerators
        accelGroup = gtk.AccelGroup()
        self.add_accel_group(accelGroup)
        self.accelGroup = accelGroup
        accelGroup.connect_group(gtk.keysyms.Page_Down, \
                                  gtk.gdk.CONTROL_MASK, gtk.ACCEL_LOCKED, \
                                  self.on_key_cycle_tabs)
        accelGroup.connect_group(gtk.keysyms.Page_Up, \
                                  gtk.gdk.CONTROL_MASK, gtk.ACCEL_LOCKED, \
                                  self.on_key_cycle_tabs)
        accelGroup.connect_group(gtk.keysyms.W, \
                                  gtk.gdk.CONTROL_MASK, gtk.ACCEL_LOCKED, \
                                  self.on_key_close_tab)
        for i in range(1, 10):
            accelGroup.connect_group(gtk.keysyms._0 + i, \
                                  gtk.gdk.MOD1_MASK, gtk.ACCEL_LOCKED, \
                                  self.on_key_change_tab)
        
        self.conversation = conversation
        self.conversation.setIsCurrent(True)
        
        self.set_title(self.conversation.getTitle(False,False))
        self.set_role('chat')

        self.textTagCount = 0

        self.user = self.conversation.getUser()
        self.conversations = {}

        self.menu = ConversationWindowMenu(self, controller)
        
        self.vbox = gtk.VBox()
        self.tabs = gtk.Notebook()
        try:
            self.tabs.connect("page-reordered", self.onPageReordered)
        except TypeError:
            # gtk 2.8
            pass
        
        self.tabs.set_show_border(False)
        self.tabs.set_property('tab-vborder', 0)

        self.tabs.set_show_tabs(False) #ok, ok, never show tabs on window init
        self.tabs.set_scrollable(True)
        self.tabs.popup_enable()
        self.openTab(conversation)

        self.vbox.pack_start(self.menu, False, False)
        self.vbox.pack_start(self.tabs, True, True)
        
        if self.config.user['showMenubar']:
            self.menu.show_all()
        self.tabs.show_all()
        self.vbox.show()

        self.connect('delete-event', self.close)
        self.connect('focus-in-event', self.on_focus_in_event)
        self.switchId = self.tabs.connect('switch-page', self.changeTabs)
        self.showMenubarId = self.config.connect('change::showMenubar', \
            self.on_menubar_change)
            
        self.config.connect('change::avatarsInTaskbar', self.on_avatars_change)
        self.controller.msn.connect('display-picture-changed', self.on_displaypicture_change)
        
        self.connect('key-press-event', self.on_key_press)
        
        self.add(self.vbox)
        
        self.conversation.ui.hideUserList()
    
    def on_key_press(self, widget, event):
        if (event.state & gtk.gdk.CONTROL_MASK) and \
           (event.keyval == gtk.keysyms.Tab or \
           event.keyval == gtk.keysyms.ISO_Left_Tab):
            if event.state & gtk.gdk.SHIFT_MASK:
                self.cycleTabs(-1)
            else:
                self.cycleTabs(1)
            return True
    
    def on_focus_in_event(self, widget, event):
        self.unsetUrgency()
        self.conversation.ui.input.grabFocus()

    def on_size_alloc(self, widget, allocation):
        self.width = allocation.width
        self.height = allocation.height
        
    def on_key_close_tab(self, accelGroup, window, keyval, modifier):
        '''Catches events like Ctrl+W and closes current tab'''
        active = self.tabs.get_current_page()
        self.close_tab(self.conversation, active)
        
    def on_key_change_tab(self, accelGroup, window, keyval, modifier):
        '''Catches alt+number and shows tab number-1  '''
        pages = self.tabs.get_n_pages()
        new = keyval - gtk.keysyms._0 - 1
        if new < pages:
            self.showTab(new)
        
    def on_key_cycle_tabs(self, accelGroup, window, keyval, modifier):
        '''Catches events like Ctrl+AvPag and consequently changes current tab'''

        if not modifier == gtk.gdk.CONTROL_MASK:
            return
            
        if keyval == gtk.keysyms.Page_Down:
            self.cycleTabs(1)
        elif keyval == gtk.keysyms.Page_Up:
            self.cycleTabs(-1)
    
    def on_menubar_change(self, config, value, oldValue):
        '''handler for showMenubar config change'''
        if value == '1':
            self.menu.show_all()
        else:
            self.menu.hide()

    def on_avatars_change(self, config, value, oldValue):
        '''handler for avatarsInTaskbar config change'''
        self.set_icon(self.conversation.getIcon())

    def on_displaypicture_change( self, msnp, switchboard, msnobj, email):
        if self.conversation.getSwitchboard().firstUser == email:
            self.set_icon(self.conversation.getIcon())

    def cycleTabs(self, cycle):
        notebook = self.tabs
        pages = notebook.get_n_pages()
        active = notebook.get_current_page()
        self.showTab((active + cycle) % pages)
        
    def showTab( self, num ):
        '''put the num tab as the selected one'''
        
        # TODO: put page as second parameter, we currently dont use it
        # but if you do please put it here.
        self.changeTabs(self.tabs, None, num)
        self.tabs.set_current_page(num)

    def doIsTyping(self, mail):
        self.conversation.receiveTyping(mail)
    
    def getTabPage(self, conversation):
        '''return a tab page according to its conversation object'''
        
        tabNum = None

        for k, v in self.conversations.items():
            if v == conversation:
                tabNum = k
                break
                
        return tabNum
    
    def getOutputTextView(self):
        '''return the current textview'''
        #XXX
        return self.conversation.ui.textview
    
    def changeTabs(self, notebook, page, page_num):
        '''callback for the switch-page signal'''
        
        self.conversation.setIsCurrent(False)
        self.conversation = self.conversations[page_num]
        
        self.set_icon(self.conversation.getIcon())
           
        self.output = self.conversation.ui.textview
        title = self.conversation.getTitle(False,False)
        if not title == '':
            self.set_title(title)
        self.conversation.setIsCurrent(True)
        self.conversation.ui.update()
    
    def openTab(self, conversation):
        '''open a new tab, that represent a new conversation'''
        ui = conversation.ui
        tabWidget = ui.tabWidget
                
        eventbox = gtk.EventBox()
        eventbox.set_visible_window(False)
        eventbox.add(tabWidget)
        eventbox.connect('event', self.tabs_event, ui)

        tabNum = self.tabs.append_page(child=ui, tab_label=eventbox)

        #self.tabs.set_tab_reorderable(self.tabs.get_nth_page(tabNum), True)
        self.conversations[ tabNum ] = conversation
        #add the new tab label to notebook menu popup
        self.addOrUpdateMenuPopupItem(tabNum, conversation)

        conversation.tabNum = tabNum
        conversation.ui.setCloseTabCallback(self.close_tab)
        # expand and fill
        self.tabs.set_tab_label_packing(ui, True, True, gtk.PACK_START)
        # ok, ok, show tabs only if needed
        self.tabs.set_show_tabs(len(self.conversations) > 1)
        
        #conversation.ui.hideUserList()
        if not self.is_active() or len(self.conversations) == 0:
            self.tabs.set_current_page(tabNum)
            self.output = ui.textview
            self.present()
    
    def addOrUpdateMenuPopupItem(self, tabNum, conversation):
        '''add the new tab label to notebook popup menu'''
        
        page = self.tabs.get_nth_page(tabNum)
        hbox = gtk.HBox()
        hbox.set_spacing(6)
        image = gtk.Image()
        image.set_from_pixbuf(self.controller.theme.getImage('online'))
        hbox.pack_start(image, False, False)
        hbox.pack_start(gtk.Label(conversation.getTitle()))
        hbox.show_all()
        self.tabs.set_menu_label(page, hbox)

    def close_tab(self, conversation, num):
        '''Close the num-th tab'''

        self.tabs.remove_page(num)

        #erase the conversation
        self.conversations[ num ].setStatus('closed')
        self.conversations[ num ].setClosed(True)
        self.conversations[ num ].ui.close()
        del self.conversations[ num ]

        # compact the conversations because we get a hole now
        keyList = [ x for x in self.conversations.keys() ]
        keyList.sort()

        for i in range(0, self.tabs.get_n_pages()):

            key = keyList[ 0 ]
            if len(keyList) > 1:
                keyList = keyList[ 1: ]

            self.conversations[ i ] = self.conversations[ key ]
            self.conversations[ i ].tabNum = i
            if i != key:
                del self.conversations[ key ]

        if(len(self.conversations) == 0):
            self.destroy()
        else:
            self.conversation = self.conversations[ self.tabs.get_current_page() ]
            self.set_title(self.conversation.getTitle(False,False))
            self.conversation.setIsCurrent(True)
            self.output = self.conversation.ui.textview
            self.tabs.set_show_tabs(len(self.conversations) > 1)
    
    def tabs_event(self, widget, event, ui):
        if event.type == gtk.gdk.BUTTON_PRESS and event.button == 2:
            self.close_tab(self.conversation, ui.parentConversation.tabNum)
    
    def clearOutputText(self, *args):
        self.conversation.textBuffer.set_text('')

    def show(self):
        gtk.Window.show(self)
        self.notOpen = False

    def close(self, *args):
        '''close the window and set the attribute self.closed to true'''

        self.hide()
        
        for i in self.conversations.keys():
            self.conversations[ i ].setClosed(True)

        self.closed = True
        
        self.controller.config.user['convWindowWidth'] = self.width
        self.controller.config.user['convWindowHeight'] = self.height

        self.config.disconnect(self.showMenubarId)
        self.tabs.disconnect(self.switchId)

        return True

    def isClosed(self):
        '''return True if the window has been hided'''
        return self.closed

    def scrollToBottom(self, tab_num = None):
        '''scroll to the end of the conversation the textview in tha tab_num
        if None scroll the actual'''

        if tab_num == None:
            textview = self.output
        else:
            textview = self.conversations[ tab_num ].ui.textview

        textbuffer = textview.get_buffer()
        mark = textbuffer.create_mark("end", textbuffer.get_end_iter(), False)
        if self.last_mark:
            y = textview.get_iter_location( textbuffer.get_iter_at_mark(self.last_mark) ).y
            pos = textview.get_visible_rect().y + textview.get_visible_rect().height 
            if pos+25 < y:
                return

        self.last_mark = mark
        textview.scroll_to_mark(mark, 0.05, True, 0.0, 1.0)
        textbuffer.place_cursor( textbuffer.get_end_iter() )

    def addChildAtAnchor(self, child, anchor, tab_num = None):
        '''add a child to an anchor in the current output if tab_num is None, otherwise add it in the tab'''

        if tab_num == None:
            self.output.add_child_at_anchor(child, anchor)
        else:
            self.conversations[ tab_num ].ui.textview.add_child_at_anchor(child, anchor)

    def setUrgency(self):
        '''try to set the urgency hint to the window'''

        if not self.has_toplevel_focus():
            try:
                self.set_urgency_hint(True)
            except:
                pass

    def unsetUrgency(self):
        '''try to unset the urgency hint to the window'''

        try:
            self.set_urgency_hint(False)
        except:
            pass
        
        self.conversation.ui.update()
        
    def onPageReordered(self, notebook, child, page_num):
        '''re-set tabNum value for every conversation''' 
        for page_num in range(notebook.get_n_pages()):
            notebook.get_nth_page(page_num).tabWidget.parentUI.parentConversation.tabNum = page_num
        
    def changeFont(self):
        '''opens the font selection dialog'''
        fontDesc = pango.FontDescription( self.config.user['fontFace'] )
        if self.config.user['fontItalic']:
            uStyle = pango.STYLE_ITALIC
        else:
            uStyle = pango.STYLE_NORMAL
        
        if self.config.user['fontBold']:
            uWeight = pango.WEIGHT_BOLD
        else:
            uWeight = pango.WEIGHT_NORMAL
        
        fontDesc.set_style( uStyle )
        fontDesc.set_weight( uWeight )
        fontDesc.set_size( self.config.user['fontSize'] * pango.SCALE )

        fontDialog = gtk.FontSelectionDialog( _( 'Choose a font' ) )
        fontDialog.set_font_name( fontDesc.to_string() )
        response = fontDialog.run()

        if response == gtk.RESPONSE_OK:
            pangoDesc = pango.FontDescription( fontDialog.get_font_name() )
            font = pangoDesc.get_family()
            style = pangoDesc.get_style()
            weight = pangoDesc.get_weight()

            italic = bold = False

            if style == pango.STYLE_ITALIC:
                italic = True
            if weight == pango.WEIGHT_BOLD:
                bold = True

            self.controller.emit( 'font-changed', font, bold, italic, pangoDesc.get_size() / pango.SCALE )
            
        fontDialog.destroy()

    def changeColor(self):
        '''opens the color selection dialog'''
        colorDialog = gtk.ColorSelectionDialog(_('Choose a color'))
        colorDialog.colorsel.set_has_palette(True)
        colorDialog.colorsel.set_current_color( \
            gtk.gdk.color_parse(self.config.user['fontColor']) )
        response = colorDialog.run()

        if response == gtk.RESPONSE_OK:
            color = colorDialog.colorsel.get_current_color()
            
            red = color.red >> 8
            green = color.green >> 8
            blue = color.blue >> 8
            
            self.controller.emit('color-changed', \
                '#%02X%02X%02X' % (red, green, blue) )
            
        colorDialog.destroy()
        
    def getConversation(self):
        '''returns the current conversation'''
        return self.conversation

class ConversationWindowMenu(gtk.MenuBar):
    '''This class represent the menu in the conversation window, i define
    it here because no other module here cares about this one.'''

    def __init__(self, parentConversationWindow, controller):
        '''Contructor'''

        gtk.MenuBar.__init__(self)

        self.parentConversationWindow = parentConversationWindow
        self.controller = controller
        accelGroup = parentConversationWindow.accelGroup

        # --- File menu
        conversationMenu = gtk.Menu()
        conversationMenuItem = gtk.MenuItem(_("_File"))
        conversationMenuItem.set_submenu(conversationMenu)

        quitMenuItem = gtk.ImageMenuItem(gtk.STOCK_QUIT)
        quitMenuItem.add_accelerator('activate', accelGroup, ord('Q'), gtk.gdk.CONTROL_MASK, gtk.ACCEL_VISIBLE)
        
        quitMenuItem.connect('activate', self.onQuitActivate)
        
        conversationMenu.add(quitMenuItem)
        self.add(conversationMenuItem)       
        
        # --- Format menu
        formatMenu = gtk.Menu()
        formatMenuItem = gtk.MenuItem(_("For_mat"))
        formatMenuItem.set_submenu(formatMenu)
        
        fontMenuItem = gtk.ImageMenuItem(gtk.STOCK_SELECT_FONT)
        colorMenuItem = gtk.ImageMenuItem(gtk.STOCK_SELECT_COLOR)

        fontMenuItem.connect('activate', self.onFontActivate)
        colorMenuItem.connect('activate', self.onColorActivate)
        
        formatMenu.add(fontMenuItem)
        formatMenu.add(colorMenuItem)
        self.add(formatMenuItem)

    def onQuitActivate(self, *args):
        '''This methd is called when Quit  is activated on the menu'''
        self.parentConversationWindow.close()

    def onLogActivate(self, check):
        '''This methd is called when Log is activated on the menu'''
        self.parentConversationWindow.conversation.setDoLog(check.get_active())

    def onFontActivate(self, *args):
        '''This method is called when font is activated on the menu'''
        self.parentConversationWindow.changeFont()

    def onColorActivate(self, *args):
        '''This method is called when color is activated on the menu'''
        self.parentConversationWindow.changeColor()

