# -*- coding: utf-8 -*-

#   This file is part of emesene.
#
#    Emesene is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    emesene is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with emesene; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import os
import time
import gtk
import gobject

import emesenelib.common
import desktop

class FileTransfer:
    ''' this class represents one filetransfer'''
    
    def __init__(self, controller, p2p, conversation, session, context, sender):
        #constants
        #self.state should always be one of those states
        self.WAITING = 0
        self.TRANSFERING = 1
        self.RECEIVED = 2
        self.FAILED = 3
        
        self.controller = controller
        self.p2p = p2p
        self.conversation = conversation
        self.session = session
        self.context = context
        self.sender = sender
        
        self.receivedBytes = 0
        self.state = self.WAITING
        
        self.timeAccepted = None
        
        if context.preview != '':
            loader = gtk.gdk.PixbufLoader()
            loader.write(context.preview)
            loader.close()
            
            self.previewImage = loader.get_pixbuf()
            loader = None
        else:
            self.previewImage = None
        
        self.localPath = None

        args = {'mail': self.p2p.mail, 'file': self.context.filename}
        self.conversation.appendOutputText(None, \
            _('%(mail)s is sending you %(file)s') % args, \
            'information')

        self.p2p.connect( 'file-transfer-progress', self.onFtProgress )
        self.p2p.connect( 'file-transfer-failed', self.onFtFailed )
        self.p2p.connect( 'file-transfer-received', self.onFtReceived )
    
    def getPreviewImage(self):
        return self.previewImage
    
    def getBytes(self):
        '''returns a tuple with received and total bytes'''
        return int(self.receivedBytes), int(self.context.file_size)
    
    def accept(self):
        '''accept this transfer'''

        self.conversation.appendOutputText(None, \
            _('You have accepted %s') % self.context.filename, \
            'information')

        self.p2p.emit('file-transfer-accepted', self.session, self.context, self.sender)
        self.timeAccepted = time.time()
        
        self.state = self.TRANSFERING
        self.ui.stateChanged()
        
    def cancel(self):
        '''cancel this transfer'''
        # FIXME: implement canceling of transfer in emesenelib
        pass
    
    def getElapsedTime(self):
        '''return the time elapsed since this transfer was accepted'''
        if not self.timeAccepted:
            return 0
        else:
            return int(time.time()) - int(self.timeAccepted)
    
    def getAverageSpeed(self):
        '''get average file transfer speed (bytes per second)'''
        time = self.getElapsedTime()
        if time == 0: # prevent division by zero
            return 0
        else: 
            return self.receivedBytes / time
    
    def remove(self):
        '''remove this transfer'''
        self.conversation.transfers.remove(self)
      
    def onFtProgress(self, switchboard, session, context, bytes, sender):
        if context != self.context: return

        self.receivedBytes = int(bytes)
        self.ui.updateProgress()

    def onFtFailed(self, switchboard, session, context, reason, sender):
        if context != self.context: return
        
        output = _('Transfer of %s failed.') % context.filename + ' ' 
        if reason == 'corrupt':
            output += _('Some parts of the file are missing')
        elif reason == 'cancelled':
            output += _('Interrupted by user')
        self.conversation.appendOutputText(None, output, 'error')

        self.state = self.FAILED
        self.ui.stateChanged()
        
    def onFtReceived(self, switchboard, session, context, buffer, sender):
        '''called when filetransfer is finished'''
        if context != self.context: return
        
        self.conversation.appendOutputText(None, \
            _('%s received successfully.') % context.filename, 'information')

        # get file dir
        config = self.controller.config
        receivedFilesDir = os.path.expanduser(config.user['receivedFilesDir'])
        
        # if directory is invalid, save to home directory
        if not os.path.exists(receivedFilesDir):
            print receivedFilesDir + ' does not exist. Saving files to home directory.'
            receivedFilesDir = os.path.expanduser('~/')
        
        name = os.path.join(receivedFilesDir, self.getFilename())
        
        num = 0
        numstr = ''
        
        while os.path.exists(name + numstr):
            num += 1
            numstr = '.' + str(num)

        self.localPath = name + numstr
        
        rc = open(self.localPath, 'wb')
        rc.write(buffer.data)
        rc.close()
        
        self.state = self.RECEIVED
        self.ui.stateChanged()
    
    def open(self):
        '''open received file'''
        if self.localPath != None:
            desktop.open(self.localPath)
    
    def getFilename(self):
        return self.context.filename
    
    def setBytesReceived(self, bytes):
        self.receivedBytes = bytes
        
    def getFraction(self):
        '''return the progress of this transfer as a float from 0 to 1'''
        return float(self.receivedBytes) / float(self.context.file_size)
