# -*- coding: utf-8 -*-

#   This file is part of emesene.
#
#    Emesene is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    emesene is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with emesene; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import sys
import gtk
import gtk.gdk
import pango
import cairo
import gobject
from Parser import Smiley
from emesenelib.ContactData import Contact, Group
from gobject import PARAM_READWRITE, TYPE_PYOBJECT

class SmileyRenderer(gtk.GenericCellRenderer):
    '''Cellrenderer with smiley support. '''

    __gproperties__ = {
        'markup': (TYPE_PYOBJECT, 'Smiley markup','A list with strs, Smileys, and \ns', PARAM_READWRITE),
        'obj': (TYPE_PYOBJECT, 'Smiley markup','A list with strs, Smileys, and \ns', PARAM_READWRITE),
        'ellipsize': (bool , 'Ellipsize True/False', '',False, gobject.PARAM_READWRITE),
        'wrap': (bool , 'Wrap True/False', '',False, gobject.PARAM_READWRITE),
    }

    _markup = ['']
    _ellipsize = True
    _wrap = True
    _smiley_layout = None    
    _obj_cached_markup = {}
    _obj_cached_layout = {}    
    
    def __init__( self, cache = {} ):
        self.__gobject_init__()
        self.xpad = 2
        self.ypad = 1
        #self._smiley_layout= SmileyLayout(self.create_pango_context())
        
    def do_get_property(self, prop):
        return getattr(self, prop.name)

    def do_set_property(self, prop, value):
        setattr(self, prop.name, value)

    def on_render(self, win, widget, background_area, cell_area, expose_area, flags):
        '''Called by gtk to render the cell.'''
        cell_x, cell_y, cell_width, cell_height = cell_area
        layout = self._obj_cached_layout[self.__get_identifier()]
        layout.set_width((cell_width - self.xpad)  * pango.SCALE)
        if flags in (int(gtk.CELL_RENDERER_SELECTED), int(gtk.CELL_RENDERER_SELECTED) + int(gtk.CELL_RENDERER_PRELIT) ):
            layout.set_in_color_override_mode(True)
        else:
            layout.set_in_color_override_mode(False)
        ctx = win.cairo_create()
        x, y = cell_x + self.xpad, cell_y + self.ypad 
        ctx.rectangle(x, y , cell_width, cell_height)
        ctx.clip()
        layout.draw(ctx, x, y)
        
    def on_get_size(self, widget, cell_area):
        '''Returns the size of the cellrenderer'''
        indentifier = self.__get_identifier()
        if self._obj_cached_markup.has_key(indentifier) and self._obj_cached_markup[indentifier] == self.markup:
            layout = self._obj_cached_layout[indentifier]
        else:
            layout = SmileyLayout(widget.create_pango_context(), self.markup) 
            layout.set_ellipsize(pango.ELLIPSIZE_END)
            self._obj_cached_markup[indentifier] = self.markup
            self._obj_cached_layout[indentifier] = layout
            style = widget.get_style()
            layout.set_override_color(style.text[gtk.STATE_SELECTED])
            #layout.set_in_color_override_mode
        width, height = layout.get_pixel_size()
        return (0,0, -1, height + (self.ypad * 2))

    def __get_identifier(self):
        if type(self.obj) == Contact: indentifier = self.obj.email
        elif type(self.obj) == Group: indentifier = self.obj.id
        else: indentifier = ''
        return  indentifier

class SmileyLabel(gtk.Widget):
    '''Label with smiley support. '''

    __gsignals__ = { 'size_request' : 'override','size-allocate':'override', 'expose-event' : 'override'}

    _text = ['']     
    _ellipsize = True
    _wrap = True
    _smiley_layout = None
            
    def __init__(self, smiley_theme):
        gtk.Widget.__init__(self)
        self.set_flags(self.flags() | gtk.NO_WINDOW)
        self._smiley_layout= SmileyLayout(self.create_pango_context())

    def set_ellipsize(self, ellipsize):
        ''' Sets the ellipsize behavior '''
        self._ellipsize = ellipsize   
        self.queue_resize()

    def set_wrap(self, wrap):
        ''' Sets the wrap behavior '''
        self._wrap = wrap
        self.queue_resize()

    def set_markup( self, text=['']):
    	self.set_text( text )
    	        
    def set_text(self, text=['']):
        ''' Sets widget text '''
        self._text = text
        self.setup_smiley_layout()
        self.queue_resize()
                
    def set_smiley_scaling(self, smiley_scaling):
        self._smiley_layout.set_smiley_scaling(smiley_scaling)
        self.queue_resize()
        
    def setup_smiley_layout(self):
        self._smiley_layout.set_element_list(self._text)

    def do_realize(self):
        gtk.Widget.do_realize(self)
        self.set_flags(self.flags() | gtk.REALIZED)
        self.window = self.get_parent().window

    def do_size_request(self, requisition):             
        self._smiley_layout.set_width(-1)
        width, height = self._smiley_layout.get_pixel_size()
        requisition.height = height
        if self._ellipsize or self._wrap:
            requisition.width = 0
        else: 
            requisition.width = width

    def do_size_allocate(self, allocation):
        if not (self._ellipsize or self._wrap):
            self._smiley_layout.set_width(-1)
            width, height = self._smiley_layout.get_pixel_size()
            self.set_size_request( width, height)
        else:
            if self._ellipsize: 
                self._smiley_layout.set_ellipsize(pango.ELLIPSIZE_END)
            else: 
                self._smiley_layout.set_ellipsize(pango.ELLIPSIZE_NONE)
            self._smiley_layout.set_width(allocation.width * pango.SCALE)
            self.set_size_request(-1, self._smiley_layout.get_pixel_size()[1] )
        gtk.Widget.do_size_allocate(self, allocation)
                        
    def do_expose_event(self, event):
        area = self.get_allocation()        
        ctx = event.window.cairo_create()
        ctx.rectangle(area.x, area.y  , area.width, area.height)
        ctx.clip()        
        self._smiley_layout.draw(ctx, area.x, area.y)
                
gobject.type_register(SmileyLabel)


class SmileyLayout(pango.Layout):

    _width = -1
    _text = ''                          # plain text
    _base_attrlist = None               # attr list without shape attributes    
    _attrlist = None                    # attr list with shape attributes    
    _override_base_attrlist = None      # override attr list without shape attributes    
    _override_attrlist = None           # override attr list with shape attributes
    _cache_ellipsize_attrlist = {}
    _in_override = False                # color override mode, used for selected text
    _override_color = gtk.gdk.Color()   
    _smilies = {}                       # key:(int"index position in the layout"), value(pixbuf"smiley")
    _smiley_info = {}                   # key:(int"index position in the layout"), value(width, height, cairoscale)
    _ellipsize_layout = None
    _scaling = -1                       # relative to ascent + desent,.. -1 for natural size
    _base_to_center = 0
    _text_height = 0
    _ellipsize = True

    def __init__(self, context, parsed_elements_list = [''], scaling=1.0):
        pango.Layout.__init__(self, context)
        self._scaling = scaling
        self.__update_base(parsed_elements_list)
        self.__update_smilies()
        self.__update_attributes()
        self.__update_layout()
        self._ellipsize_layout = pango.Layout(context)
        self._ellipsize_layout.set_text('___')
        self._ellipsize_layout.set_ellipsize(pango.ELLIPSIZE_END)
        self._ellipsize_layout.set_width(0)

    def set_element_list(self, parsed_elements_list=['']):
        self.__update_base(parsed_elements_list)
        self.__update_smilies()
        self.__update_attributes()
    
    def set_text(self, text):
        if type(text) in (str, unicode):
            self.set_element_list([text])
    
    def set_markup(self, markup):
        self.set_text(markup)

    def get_size(self):
        natural_width, natural_height = pango.Layout.get_size(self)        
        if self._width >= 0 and self._ellipsize : # if ellipsize
            return self._width, natural_height
        else: 
            return natural_width, natural_height
        
    def get_pixel_size(self):
        natural_width, natural_height = pango.Layout.get_pixel_size(self)        
        if self._width >= 0 and self._ellipsize : # if ellipsize
            return pango.PIXELS(self._width), natural_height
        else: 
            return natural_width, natural_height    
        
    def set_width(self, width):
        self._width = width
        self.__update_layout()
        
    def get_width(self):
        return self._width 
           
    def set_ellipsize(self, value):
        if value == pango.ELLIPSIZE_END:
            self._ellipsize = True
        else:
            self._ellipsize = False
        self.__update_layout()
        
    def set_smiley_scaling(self, smiley_scaling):
        self._scaling = smiley_scaling
        self.__update_smilies()
        self.__update_attributes()
        
    def set_in_color_override_mode(self, in_override):
        if not in_override == self._in_override:
            self._in_override = in_override
            self.__update_attributes()

    def set_override_color(self, gdk_color):
        if type(gdk_color) == gtk.gdk.Color:
            self._override_color = gdk_color   
            self._override_attrlist = self.__create_override_attrlist(self._override_attrlist)
            self._override_base_attrlist = self.__create_override_attrlist(self._override_base_attrlist)
            self.__update_attributes()

    def __update_base(self, parsed_elements_list=['']):
        self._smilies = {}
        self._base_attrlist = pango.AttrList()
        self._text = ''
        stack = []
        attrs = []
        index = 0
        if type(parsed_elements_list) in (str, unicode): parsed_elements_list = [parsed_elements_list]
        for element in parsed_elements_list:
            if type(element) in (str, unicode):
                try:
                    attrlist, text, accel = pango.parse_markup(str(element), u'\x00')
                except:
                    print 'SMIELY RENDERER has been give non valid pango markup!'
                    print str(element)
                    attrlist, text = pango.AttrList(), str(element)
                self.__combine_attrlist(self._base_attrlist, attrlist, len(self._text))
                index += len(text)                
                self._text += text
            elif type(element) == Smiley:
                pix = element.getPixbuf(  animated=False )
                self._smilies[index] = pix                    
                self._text += '_'
                index += 1
        self._override_base_attrlist = self.__create_override_attrlist(self._base_attrlist)
        pango.Layout.set_text(self, self._text)
        logical = self.get_line(0).get_pixel_extents()[1]
        ascent = pango.ASCENT(logical)
        decent = pango.DESCENT(logical)
        self._text_height =  ascent + decent      
        self._base_to_center = (self._text_height / 2) - decent
        
    def __combine_attrlist(self, base_attrlist, attrlist, shift):
        itter = attrlist.get_iterator()
        next = True
        while next:
            attrs = itter.get_attrs()
            for attr in attrs:
                attr.end_index += shift
                attr.start_index += shift
                base_attrlist.insert(attr)
            next = itter.next()

    def __create_override_attrlist(self, attrlist):
        color = self._override_color
        new_list = pango.AttrList()
        new_list.insert(pango.AttrForeground(color.red, color.green, color.blue, 0, len(self._text)))
        itter = attrlist.get_iterator()
        next = True
        while next:
            attrs = itter.get_attrs()
            for attr in attrs:
                if not (attr.type in (pango.ATTR_FOREGROUND, pango.ATTR_BACKGROUND)):
                    new_list.insert(attr.copy())
            next = itter.next()
        return new_list
        
    def __update_smilies(self):
        self._attrlist = self._base_attrlist.copy()
        self._smiley_info = {}
        #set max height of a pixbuf
        if self._scaling >= 0: max_height = self._text_height * self._scaling
        else: max_height = sys.maxint                
        for index, pixbuf in self._smilies.iteritems():
            height, width = pixbuf.get_height() , pixbuf.get_width()
            npix = pixbuf.copy()
            if height > max_height:
                cairo_scale = float(max_height) / float(height)
                height = int(height * cairo_scale)
                width = int(width * cairo_scale)
                npix = npix.scale_simple(width, height, gtk.gdk.INTERP_BILINEAR)
            self._smiley_info[index] = npix
            rect = (0, -1 * (self._base_to_center + (height /2)) * pango.SCALE, width * pango.SCALE, height * pango.SCALE)
            self._attrlist.insert(pango.AttrShape( (0,0,0,0), rect, index, index + 1 ))    
        self._override_attrlist = self.__create_override_attrlist(self._attrlist)
        
    def __update_attributes(self):
        if self._in_override: self.set_attributes(self._override_attrlist)
        else: self.set_attributes(self._attrlist)
    
    def __update_layout(self):
        if self._width >= 0 and self._ellipsize == False: # if true , then wrap
            pango.Layout.set_width(self, self._width)
        else:
            pango.Layout.set_width(self, -1)
   
    def draw(self, ctx, x, y):
        ctx.translate( x,y)
        #Clipping and ellipsation
        if self._width >= 0:        
            INLINE, BYTE, GRAPH = 0, 1, 2 
            X,Y,WIDTH,HEIGHT = 0, 1, 2, 3       
            layout_width = pango.PIXELS(self._width)
            coordinates = [] #of path in px
            for i in range(self.get_line_count()):
                line = self.get_line(i)
                edge = line.x_to_index(layout_width * pango.SCALE)        
                if edge[INLINE]: 
                    #create ellipsize layout with the style of the char
                    self._ellipsize_layout.set_attributes(self.__get_ellipsize_attrlist(edge[BYTE]))
                    ellipsize_width = self._ellipsize_layout.get_pixel_size()[0]           
                    edge = line.x_to_index((layout_width * pango.SCALE) - (ellipsize_width * pango.SCALE))
                    char = self.index_to_pos(edge[BYTE])
                    if edge[INLINE]: # use the x position here and its height
                        x1, y1 = pango.PIXELS(char[X]), pango.PIXELS(char[Y])
                        x2, y2 = pango.PIXELS(char[X]), pango.PIXELS(char[Y] + char[HEIGHT])
                    else: # x position is now 0 , char height
                        x1, y1 = pango.PIXELS(0), pango.PIXELS(char[Y])
                        x2, y2 = pango.PIXELS(0), pango.PIXELS(char[Y] + char[HEIGHT])                
                    coordinates.append((x1, y1))
                    coordinates.append((x2, y2))                           
                    ellipsize_ascent = pango.ASCENT(self._ellipsize_layout.get_line(0).get_pixel_extents()[1])                    
                    line_ascent = pango.ASCENT(line.get_pixel_extents()[1])                                      
                    ctx.move_to(x1, y1 + line_ascent - ellipsize_ascent)
                    ctx.show_layout(self._ellipsize_layout)
                else:
                    char = self.index_to_pos(edge[BYTE])
                    coordinates.append((pango.PIXELS(char[X] + char[WIDTH]), pango.PIXELS(char[Y])))
                    coordinates.append((pango.PIXELS(char[X] + char[WIDTH]), pango.PIXELS(char[Y] + char[HEIGHT])))
            if len(coordinates)> 0:
                ctx.move_to(0, 0)
                for x, y in coordinates:
                    ctx.line_to(x, y)
                ctx.line_to(0, coordinates[-1][1])
                ctx.close_path()
                ctx.clip()
        #layout
        ctx.move_to(0,0)
        ctx.show_layout(self)
        #smilies
        for index  in self._smilies.keys():
            x, y, width, height = self.index_to_pos(index)                        
            pixbuf = self._smiley_info[index] 
            tx, ty = (pango.PIXELS(x), pango.PIXELS(y) + (pango.PIXELS(height)/2) - (pixbuf.get_height() / 2))
            ctx.set_source_pixbuf(pixbuf, tx, ty)
            ctx.paint()

    def __get_ellipsize_attrlist(self, index):
        if self._in_override: itter = self._override_base_attrlist.get_iterator()
        else: itter = self._base_attrlist.get_iterator()
        attrlist = pango.AttrList()
        next = True
        while next:
            attrs = itter.get_attrs()
            for attr in attrs:
                start, end = itter.range()
                if start <= index < end:
                    n_attr = attr.copy()
                    n_attr.start_index = 0                
                    n_attr.end_index = 3
                    attrlist.insert(n_attr)
            next = itter.next()
        return attrlist
