# -*- coding: utf-8 -*-

#   This file is part of emesene.
#
#    Emesene is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    emesene is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with emesene; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import base64
import gobject

import common
import msn_p2p

class P2PManager(dict):
    '''global p2p manager'''

    def __init__(self, msn):
        self._users = {}
        self.msn = msn

    def __repr__(self):
        return '<P2PManager with ' + len(self._users) + ' users>'

    def __getitem__(self, mail):
        mail = mail.lower()
        if mail not in self._users:
            self._users[mail] = P2PUser(self, mail)
        return self._users[mail]
    
    def __setitem__(self, mail, value):
        raise TypeError('Read only')

class P2PUser(gobject.GObject):
    '''this class manage the creation of objects to handle the p2p
    messages and connect the necessary signals'''

    __gsignals__ = {
        'invite-message-received': (gobject.SIGNAL_RUN_LAST, 
            gobject.TYPE_NONE,(gobject.TYPE_PYOBJECT,)),

        # BinHeader, dict, message
        'msnp2p-message-received': (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE,
         (gobject.TYPE_PYOBJECT, gobject.TYPE_PYOBJECT, gobject.TYPE_PYOBJECT)),
       
        # this signal is emited when the actual image is received
        # msnobj, str, sender
        'custom-emoticon-data-received': (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE,
         (gobject.TYPE_PYOBJECT, gobject.TYPE_PYOBJECT, gobject.TYPE_PYOBJECT)),
        
        # msnobj, str, sender
        'display-picture-received': (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE,
         (gobject.TYPE_PYOBJECT, gobject.TYPE_PYOBJECT, gobject.TYPE_PYOBJECT)),
        
        # session, context, sender 
        'file-transfer-invite': (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE,
         (gobject.TYPE_PYOBJECT, gobject.TYPE_PYOBJECT, gobject.TYPE_PYOBJECT)),
        
        # session, context, sender 
        'file-transfer-accepted': (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE,
         (gobject.TYPE_PYOBJECT, gobject.TYPE_PYOBJECT, gobject.TYPE_PYOBJECT)),

        # session, context, buffer, sender 
        'file-transfer-received': (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE,
         (gobject.TYPE_PYOBJECT, gobject.TYPE_PYOBJECT, \
          gobject.TYPE_PYOBJECT, gobject.TYPE_PYOBJECT)),
        
        # session, context, reason, sender
        'file-transfer-failed': (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE,
        (gobject.TYPE_PYOBJECT, gobject.TYPE_PYOBJECT, \
         gobject.TYPE_STRING, gobject.TYPE_PYOBJECT)),

        # session, context, bytes, sender
        'file-transfer-progress': (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE,
        (gobject.TYPE_PYOBJECT, gobject.TYPE_PYOBJECT, \
         gobject.TYPE_STRING, gobject.TYPE_PYOBJECT)),
    }

    def __init__(self, manager, mail):
        '''the switchboard is the one this manager is attached, in
        this way if we have too many open conversations, the amount
        of handlers that inspect a messager are less'''

        gobject.GObject.__init__( self )

        self.manager = manager
        self.mail = mail
        self.object_manager = manager.msn.msnObjectsManager

        self.transport = None
        self.switchboard = None
        self.direct = None

        self.connect('display-picture-received', \
            self.manager.msn.on_display_picture_received)
        self.connect('custom-emoticon-data-received', \
            self.manager.msn.on_custom_emoticon_received)
        
        # TODO: move the partial invite handling code from switchboard
        #pending here means that the whole message hasn't been received yet
        # format: {identifier: [message, ...] }
        self.pending_invites = {}
    
    def __repr__(self):
        return "<P2P User: " + self.mail + ">"
    
    def do_invite_message_received(self, message):
        '''method called when an invite message is received
        in the switchboard'''
        
        bin_header = msn_p2p.get_bin_header(message)

        if bin_header.total_data_size > bin_header.message_length:
            #partial invite

            identifier = bin_header.identifier
            if identifier in self.pending_invites:
                # follow-up message
                self.pending_invites[identifier].append(message)
                
                size = bin_header.data_offset + bin_header.message_length
                if size < bin_header.total_data_size:
                    # don't process, it's not ready yet
                    return
                
                # rebuild a full invite message
                message = ''
                for msg in self.pending_invites[identifier]:
                    tmp_header = msn_p2p.get_bin_header(msg)
                    message += msn_p2p.get_data_chunk(msg, tmp_header)

                del self.pending_invites[identifier]
            else:
                # first message
                self.pending_invites[identifier] = [message]
                return

        parameters = msn_p2p.get_parameters(message)

        if 'Context' in parameters:
            if 'EUF-GUID' not in parameters:
                common.debug('no EUF-GUID in p2p invite', 'p2p')
                return
                
            if parameters['EUF-GUID'] == msn_p2p.Base.EUFGUID_FILE:
                # is a file
                # we should ask the user here...
                rawcontext = parameters['Context']
                context = msn_p2p.FTContext(base64.b64decode(rawcontext))
                context.print_fields()
                # create a ft receiver here
                data = msn_p2p.Buffer('temp.tmp', '')
                if data.data is not None:
                    self.get_switchboard()
                    msn_p2p.FTReceiver(self, context, parameters, \
                        bin_header, data)
            else:
                data = msn_p2p.Buffer('temp.tmp', \
                    self.get_data_by_context(parameters['Context']) )
                if data.data is not None:
                    msn_p2p.Sender(self, bin_header, parameters, data)
                else:
                    common.debug('invalid context in MsnP2PManager', 'p2p')
        elif 'Conn-Type' in parameters:
            # its a direct connect invite
            data = msn_p2p.Buffer('','')
            msn_p2p.Sender(self, bin_header, parameters, data)
        else:
            common.debug('no context on invite message in MsnP2PManager', 'p2p')

    def get_data_by_context(self, context):
        '''receive a base64 encoded msnobj and try to
        return a Buffer instace with the data, if
        the msnobj doesnt exist, return None'''
        
        try:
            msnobj = self.object_manager.getByContext(context)
            try:
                return open(msnobj.filename, 'rb').read()
            except IOError:
                return None
        except TypeError, AttributeError:
            common.debug('invalid msnobj in get_data_by_context', 'p2p')
            return None
        
    def on_msnp2p_message_ready(self, base, message, dest):
        '''method called when a p2p message is ready to be sent'''

        # this function will have to choose between switchboard or direct
        # connection, and start those connections if needed
        
        self.get_switchboard()
        self.switchboard.sendMessage(message, '', 'application/x-msnmsgrp2p', \
            'D', 'P2P-Dest: ' + dest )

    def get_switchboard(self):
        if self.switchboard is None or \
           self.switchboard.status in ('closed', 'error'):
            self.switchboard = self.manager.msn.getSwitchboard(self.mail)

