# -*- coding: utf-8 -*-

#   This file is part of emesene.
#
#    Emesene is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    emesene is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with emesene; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

'''module to handle p2p sessions on msn protocol'''

import os # temp for file saving
import time
import random
import base64
import struct
import gobject

import common
import Msnobj

common.debugFlag = True

def get_bin_header(message):
    '''receive a msnp2p message and return a BinHeader
    instance with the bin header of the message
    if you send crap, you will get a nice BinHeader full
    of crap! ;).
    a message is all the p2p part except the header
    that means, no MSG, MIME-Version, P2P-Dest,
    that can be achieved whit a message.split('\r\n')[1:]'''

    return BinHeader(message[:48])

def get_parameters(message):
    '''receive a msnp2p message and return a dict with all
    the parameters in the msnlp message or an empty dict
    if no msnlp header found
    a message is all the p2p part except the header
    that means, no MSG, MIME-Version, P2P-Dest,
    that can be achieved whit a message.split('\r\n')[1:]'''
    parameters = {}

    for line in message.split('\r\n'):
        try:
            key, value = line.split(': ')
            parameters[key] = value
        except ValueError:
            pass

    return parameters

def get_from(string):
    '''return the mail contained on a from line of 
    a msnlp message, an empty string otherwise'''

    try:
        return string.split('<msnmsgr:')[1].split('>')[0]
    except IndexError:
        return ''

# the same method but with an according name
get_to = get_from
        

def get_via_branch(string):
    '''extract the via branch WITHOUT THE BRACES from a Via: 
    line, return an empty string otherwise'''

    try:
        return string.split('branch={')[1].split('}')[0]
    except IndexError:
        return ''

def get_call_id(string):
    '''try to remove the braces from a call id line, return 
    an empty string otherwise'''
    
    try:
        return string.split('{')[1].split('}')[0]
    except IndexError:
        return ''

def get_data_chunk(message, bin):
    '''return the data chunk in message, an empty string if
    somethin went bad'''

    try:
        return message[48:48 + bin.message_length]
    except IndexError:
        return ''
    
def is_invite(message):
    '''return True if the message is an invite message'''
    return contains(message, 'INVITE MSNMSGR:')

def is_dp_invite(message, params):
    '''return True if its a display picture invite'''
    return (is_invite(message) and params['EUF-GUID'] == Base.EUFGUID_DISPLAY)

def is_ce_invite(message, params):
    '''return True if its a custom emoticon invite'''
    return (is_invite(message) and params['EUF-GUID'] == Base.EUFGUID_DISPLAY)

def is_file_invite(message, params):
    '''return True if its a file invite'''
    return (is_invite(message) and params['EUF-GUID'] == Base.EUFGUID_FILE)
    
def contains(string, substring):
    '''return True if string contains substring'''
    return (string.find(substring) != -1)
    
# we use the max from sys.maxint in a 32 bit machine because
# in a 64 bit machine it is bigger and can give some troubles
def random_number(minimum = 0, maximum = 2147483647):
    '''return a random number between minimum and maximum'''
    return random.randint(minimum, maximum)
 
def random_id():
    '''return a random id according to msnp2p rules'''
    id_ = '%04X' % random.randint( 0x1111 , 0xAAFF )
    id_ += '%04X' % random.randint( 0x1111 , 0xAAFF ) + "-"
    id_ += '%04X' % random.randint( 0x1111 , 0xAAFF ) + "-"
    id_ += '%04X' % random.randint( 0x1111 , 0xAAFF ) + "-"
    id_ += '%04X' % random.randint( 0x1111 , 0xAAFF ) + "-"
    id_ += '%04X' % random.randint( 0x1111 , 0xAAFF )
    id_ += '%04X' % random.randint( 0x1111 , 0xAAFF )
    id_ += '%04X' % random.randint( 0x1111 , 0xAAFF )
    
    return id_

def make_msnlp(method, to_, from_, branch, cseq, call_id, 
    max_forwards, content_type, content_length, body, footer):
    '''create a msnlp message'''
    return Base.MSNLP_TEMPLATE % (method, to_, from_, branch, cseq, 
     call_id, max_forwards, content_type, content_length, body, footer)

def to_pretty_msnlp(message):
    '''format a msnlp message to get nice output when printing'''
    return repr(message).replace( '\\r\\n', '\\r\\n\n' )[1:-1]

def compare_binheaders(b1, b2):
    '''print the binary fields'''
    def compare_field(name, attr):
        print '%-17s %10s %10s' % (name, getattr(b1, attr), getattr(b2, attr))

    print '-------------------------------------------\n'
    compare_field('SessionID:', 'session_id')
    compare_field('Identifier:', 'identifier')
    compare_field('Data offset:', 'data_offset')
    compare_field('Total data size:', 'total_data_size')
    compare_field('Message length:', 'message_length')
    compare_field('Flag:', 'flag')
    compare_field('Ackd identifier:', 'acknowledged_identifier')
    compare_field('Ackd unique ID:', 'acknowledged_unique_id')
    compare_field('Ackd data size:', 'acknowledged_data_size')
    print '-------------------------------------------\n'

def get_dword_from_data(header, start):
    '''extract 4 bytes from header starting from start
    and return it as a number'''
    
    byte1 = ord(header[start])
    byte2 = ord(header[start + 1])
    byte3 = ord(header[start + 2]) 
    byte4 = ord(header[start + 3])
    
    return (byte4<<24) + (byte3<<16) + (byte2<<8) + byte1

def get_qword_from_data(header, start):
    '''return the qword from header, starting from start'''
    # i think its like this
    return ((get_dword_from_data(header, start + 4) << 32) + 
            get_dword_from_data(header, start))

class Base(gobject.GObject):
    '''a base class for P2P handlers (Sender, Receiver)
    this class has all the methods to create the different
    messages and to check if a message belong to a handler'''

    __gsignals__ = { 
        # message, dest
        'msnp2p-message-ready' : 
        (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, 
            (gobject.TYPE_PYOBJECT,gobject.TYPE_PYOBJECT))
    }

    # seconds alive until it set the state to STATUS_END, 
    # so it disconnect the signals
    MAX_TIME_ALIVE = 60*3

    # this status means the status we are before processing the 
    # current message you dont have to pass through all the states
    STATUS_CREATED = 0 # when just created
    STATUS_INVITE = 1 # we are on invite waiting the base id
    STATUS_BASEID = 2 # received the base id and we are waiting the 200ok
    STATUS_200OK = 3
    STATUS_200OK_ACK = 4
    STATUS_PREPARATION = 5
    STATUS_PREPARATION_ACK = 6
    STATUS_SENDING = 7
    STATUS_RECEIVING = 8
    STATUS_DATA_ACK = 9
    STATUS_BYE = 10
    STATUS_BYE_ACK = 11
    STATUS_FINISH = 12
    STATUS_END = 13 # when all the stuff is finished
    STATUS_ERROR = 14 # something gone bad :S
    
    CONTENT_LIMIT = 1202
    
    CONTENT_INVITE = 'application/x-msnmsgr-sessionreqbody'
    CONTENT_200OK = CONTENT_INVITE
    CONTENT_DECLINE = CONTENT_INVITE
    CONTENT_CLOSE = 'application/x-msnmsgr-sessionclosebody'
    CONTENT_BYE = CONTENT_CLOSE
    
    #flags

    #* 0x0 when no flags are specified
    NO_FLAG = int(0x0)
    #* 0x2 if it's an acknowledgement
    ACK_FLAG = int(0x2)
    #* 0x8 if there was an error on the binary level
    BINERROR_FLAG = int(0x8)
    #* 0x20 when the message content is data for 
    # display pictures or custom emoticons
    PIC_OR_EMO_FLAG = int(0x20)
    #* 0x40 Bye ack flag
    BYE_ACK_FLAG = int(0x40)
    # 0x80 bye flag
    BYE_FLAG = int(0x80)
    #* 0x01000030 if it's data of a file 
    FILE_FLAG = int(0x01000030)

    EUFGUID_DISPLAY = '{A4268EEC-FEC5-49E5-95C3-F126696BDBF6}'
    EUFGUID_FILE = '{5D3E02AB-6190-11D3-BBBB-00C04F795683}'
    #EUFGUID_WEBCAM = '{4BD96FC0-AB17-4425-A14A-439185962DC8}'
    
    FOOTER = '\0\0\0\0'
    DP_FOOTER = '\0\0\0\1'
    FILE_FOOTER = '\0\0\0\2'
    INK_FOOTER = '\0\0\0\3'
    CE_FOOTER = '\0\0\0\x0c'

    MSNLP_TEMPLATE = '%s\r\nTo: <msnmsgr:%s>\r\nFrom: <msnmsgr:%s>\r\n' \
    'Via: MSNSLP/1.0/TLP ;branch={%s}\r\nCSeq: %s \r\nCall-ID: {%s}'
    MSNLP_TEMPLATE += '\r\nMax-Forwards: %s\r\nContent-Type: ' \
    '%s\r\nContent-Length: %s\r\n\r\n%s%s'
        
    INVITE_BODY = 'EUF-GUID: %s\r\nSessionID: %s\r\nAppID: ' \
    '%s\r\nContext: %s\r\n\r\n\0'
    OK_BODY = 'SessionID: %s\r\n\r\n\0'
    DECLINE_BODY = 'SessionID: %s\r\n\r\n\0'
    BYE_BODY = '\r\n\0'

    INVITE_METHOD = 'INVITE MSNMSGR:%s MSNSLP/1.0'
    OK_METHOD = 'MSNSLP/1.0 200 OK'
    DECLINE_METHOD = 'MSNSLP/1.0 603 Decline'
    BYE_METHOD = 'BYE MSNMSGR:%s MSNSLP/1.0'

    def __init__(self):
        '''from_ is the sender and to the receiver email'''
        gobject.GObject.__init__(self)
        self.status = Base.STATUS_CREATED

        # the moment when its created
        self.time_stamp = time.time()

        self.from_ = ''
        self.to_ = ''
        
        # this values are just the most common ones
        self.via = random_id()
        self.cseq = '0'
        self.call_id = random_id()
        self.max_forwards = '0'
        self.content_type = ''
        self.content_length = ''
        self.euf_guid = ''
        self.session_id = str(random_number(50000))
        self.app_id = ''
        self.context = ''
        self.identifier = random_number(50000)
        self.base_id = self.identifier # base_id will be modified

        self.his_base_id = 0
        self.his_session_id = ''

        self.last_sent_header = None
        self.last_received_header = None
        self.data = None

    def check_time_limit(self):
        '''check if the time alive is greater than MAX_TIME_ALIVE
        if it is, set the status to STATUS_END so the next call
        to check, the signals get disconected and the gc delete the object'''

        if self.temporary_session and \
            time.time() > self.time_stamp + Base.MAX_TIME_ALIVE:

            common.debug('max time alive exceeded', 'p2p')
            self.status = Base.STATUS_END

    def increase_base_id(self):
        '''set the next base_id, it has a special rule
        the first time we substract 3, then we add, but if when
        adding we get the first value, we skip it'''
        if self.base_id == self.identifier:
            self.base_id -= 3
        elif self.base_id + 1 == self.identifier:
            self.base_id += 2
        else:
            self.base_id += 1
    
    
    def standard_ack(self, bin):
        '''create and return a standard ack message'''
        header = BinHeader()

        header.total_data_size = bin.total_data_size
        header.flag = Base.ACK_FLAG
        header.acknowledged_identifier = bin.identifier
        header.acknowledged_unique_id = bin.acknowledged_identifier
        header.acknowledged_data_size = bin.total_data_size

        return header

    def is_ack(self, bin, lsh=None):
        '''return True if bin is an ack message of this msnp2p session'''
        if lsh == None:
            lsh = self.last_sent_header
        
        if bin.flag == Base.ACK_FLAG and \
           lsh.identifier == bin.acknowledged_identifier and \
           lsh.acknowledged_identifier == bin.acknowledged_unique_id and \
           lsh.total_data_size == bin.acknowledged_data_size:
            return True

        return False

    # ------------------------------- INVITE -------------------------------
    # sent when i want to receive his DP or CE or when i want to send a file

    def invite_msnlp(self, context):
        '''create a invite msnlp message'''
        self.cseq = '0'
        self.max_forwards = '0'
        self.via = random_id()
        self.call_id = random_id()
        self.session_id = str(random_number(50000))
        invite_body = Base.INVITE_BODY % (Base.EUFGUID_DISPLAY, 
                        self.session_id, self.app_id, context)
        return make_msnlp(Base.INVITE_METHOD % self.to_, self.to_, 
                self.from_, self.via, self.cseq, random_id(),
                self.max_forwards, Base.CONTENT_INVITE, len(invite_body), 
                invite_body, Base.FOOTER)

    def invite_message(self, context):
        '''create the complete invite message'''
        msnlp = self.invite_msnlp(context)
        
        header = BinHeader()
        header.identifier = self.identifier
        header.acknowledged_identifier = random_number(50000)
        header.total_data_size = len(msnlp) - 4
        header.message_length = len(msnlp) - 4
        
        self.last_sent_header = header
        self.status = Base.STATUS_INVITE
        return header.data + msnlp
         
    # ------------------------------- DECLINE -------------------------------

    def decline_msnlp(self):
        '''create a decline msnlp message'''
        self.cseq = '0'
        self.max_forwards = '0'
        self.via = random_id()
        self.call_id = random_id()
        self.session_id = str(random_number(50000))
        decline_body = Base.DECLINE_BODY % (self.session_id)
        return make_msnlp(Base.DECLINE_METHOD, self.to_, 
                self.from_, self.via, self.cseq, random_id(),
                self.max_forwards, Base.CONTENT_DECLINE, len(decline_body), 
                decline_body, Base.FOOTER)

    def decline_message(self):
        '''create the complete invite message'''
        msnlp = self.decline_msnlp()
        
        header = BinHeader()
        header.identifier = self.identifier
        header.acknowledged_identifier = random_number(50000)
        header.total_data_size = len(msnlp) - 4
        header.message_length = len(msnlp) - 4
        
        self.last_sent_header = header
        self.status = Base.STATUS_END
        return header.data + msnlp
    # ---------------------------- BaseID ------------------
    # sent when a friend invite me to send him my DP or a CE

    def base_id_message(self, bin):
        '''create a base_id message'''
        header = self.standard_ack(bin)
        header.identifier = self.identifier

        self.last_received_header = bin
        self.last_sent_header = header
        self.status = Base.STATUS_BASEID
        return header.data + Base.FOOTER

    def is_base_id_message(self, bin):
        '''return True if the header belongs to this session
        and is a base id message'''
        return (self.is_ack(bin) and 
            self.identifier == bin.acknowledged_identifier)

    # ---------------------------- 200OK -------------------
    # sent after the baseid to indicate that everything is ok

    def ok_message_msnlp(self):
        '''create the msnlp header of a 200 ok message'''
        self.cseq = str(int(self.cseq) + 1)
        
        ok_body = Base.OK_BODY % (self.session_id)
        return make_msnlp(Base.OK_METHOD, self.from_, 
                self.to_, self.via, self.cseq, self.call_id, 
                self.max_forwards, Base.CONTENT_200OK,
                len(ok_body), ok_body, Base.FOOTER)

    def ok_message(self):
        '''create a 200 ok message'''
        msnlp = self.ok_message_msnlp()
        self.increase_base_id()

        header = BinHeader()
        header.identifier = self.base_id
        header.acknowledged_identifier = random_number(50000)
        header.total_data_size = len(msnlp) - 4
        header.message_length = len(msnlp) - 4
        
        self.last_sent_header = header
        self.status = Base.STATUS_200OK
        return header.data + msnlp
    
    def is_ok(self, bin, params, message):
        '''return True if the header belongs to this session
        and is a 200 ok message'''
        # WLM send bin.identifier + 4
        # bin.identifier + 3 == self.his_base_id)

        return (contains(message, 'MSNSLP/1.0 200 OK') and \
            'To' in params and get_to(params['To']) == self.from_ and \
            'From' in params and get_from(params['From']) == self.to_)

    # ----------------------------- 200OK ACK ----------------------

    def ok_ack_message(self, bin):
        '''create a 200 ok ack message'''
        self.increase_base_id()
        
        header = BinHeader()
        header.identifier = self.base_id
        header.total_data_size = bin.total_data_size
        header.acknowledged_data_size = bin.total_data_size
        header.flag = Base.ACK_FLAG
        header.acknowledged_identifier = bin.identifier
        header.acknowledged_unique_id = bin.acknowledged_identifier
        
        self.last_received_header = bin
        self.last_sent_header = header
        self.status = Base.STATUS_200OK_ACK
        return header.data + Base.FOOTER

    def is_ok_ack(self, bin):
        '''return True if the header belongs to this session
        and is a 200 ok ack message'''
       
        if self.is_ack(bin):# and \
           #self.his_base_id - 3 == bin.identifier:
            return True

        return False

    # ---------------------------- Data preparation message ---------

    def dpm_message(self, bin):
        '''create a data preparation message'''
        self.increase_base_id()
        
        header = BinHeader()
        header.session_id = int(self.his_session_id)
        header.identifier = self.base_id
        header.total_data_size = 4
        header.message_length = 4
        header.acknowledged_identifier = random_number(50000)
        
        self.last_received_header = bin
        self.last_sent_header = header
        self.status = Base.STATUS_PREPARATION
        return header.data + Base.FOOTER + Base.DP_FOOTER
        
    def is_dpm(self, bin):
        '''return True if the header belongs to this session
        and is a data preparation message'''
        if int(self.session_id) == bin.session_id and \
           bin.total_data_size == 4 and \
           bin.message_length == 4:
            return True

        return False

    # --------------------------- Data preparation message ack ------

    def dpm_ack_message(self, bin):
        '''create a data preparation ack message'''
        self.increase_base_id()
        
        header = BinHeader()
        header.session_id = int(self.session_id)
        header.identifier = self.base_id
        header.total_data_size = bin.total_data_size
        header.acknowledged_data_size = bin.total_data_size
        header.flag = Base.ACK_FLAG
        header.acknowledged_identifier = bin.identifier
        header.acknowledged_unique_id = bin.acknowledged_identifier
        
        self.last_received_header = bin
        self.last_sent_header = header
        self.status = Base.STATUS_PREPARATION_ACK
        return header.data + Base.FOOTER

    def is_dpm_ack(self, bin):
        '''return True if the header belongs to this session
        and is a data preparation ack message'''

        if self.is_ack(bin) and \
           bin.session_id == int(self.his_session_id):
            return True

        return False

    # -------------------------- Data message -----------------------

    def data_message(self):
        '''return a list of messages to send the data'''
        if not self.data or self.data.size() <= 0:
            common.debug('bad data in data_message', 'p2p')
            return []

        self.increase_base_id()
        
        data_size = self.data.size() # dont call len every iteration
        pointer = 0
        messages = []
        
        for chunk in self.data.get_chunks(Base.CONTENT_LIMIT):
            header = BinHeader()
            header.session_id = int(self.his_session_id)
            header.identifier = self.base_id
            header.data_offset = pointer
            chunk_length = len(chunk)
            pointer += chunk_length
            header.total_data_size = data_size
            header.message_length = chunk_length
            header.flag = Base.PIC_OR_EMO_FLAG
            header.acknowledged_identifier = random_number(50000)

            messages.append( header.data + chunk + Base.DP_FOOTER )
            self.last_sent_header = header

        self.status = Base.STATUS_SENDING
        return messages

    def is_data(self, bin):
        '''return True if the header is from a data message of
        this p2p session'''
        # WLM return  self.his_base_id -2
        #    bin.identifier == self.his_base_id - 1:

        if not self.is_ack(bin) and \
            bin.session_id == int(self.session_id):
            return True

        return False

    def is_last_data(self, bin):
        '''return True if it is the last data message'''

        if self.is_data(bin) and \
            bin.data_offset + bin.message_length == bin.total_data_size:
            return True

    # ------------------------- Data ack -------------------------------

    def data_ack(self, bin):
        '''return a data ack message'''
        self.increase_base_id()
        
        header = BinHeader()
        header.session_id = int(self.session_id)
        header.identifier = self.base_id
        header.total_data_size = bin.total_data_size
        header.acknowledged_data_size = bin.total_data_size
        header.flag = Base.ACK_FLAG
        header.acknowledged_identifier = bin.identifier
        header.acknowledged_unique_id = bin.acknowledged_identifier
        
        self.last_received_header = bin
        self.last_sent_header = header
        self.status = Base.STATUS_DATA_ACK
        return header.data + Base.FOOTER
        
    def is_data_ack(self, bin):
        '''return True if bin is a header od a data ack from
        this p2p session'''

        if self.is_ack(bin) and \
           bin.session_id == int(self.his_session_id):
            return True

        return False

    # -------------------------- BYE --------------------------------

    def bye_message_msnlp(self):
        '''return the msnlp part of a bye message'''
        self.cseq = '0'
        bye_body = Base.BYE_BODY
        return make_msnlp(Base.BYE_METHOD % self.from_, self.from_, 
                self.to_, random_id(), self.cseq, self.call_id,
                self.max_forwards, Base.CONTENT_BYE, len(bye_body), 
                bye_body, Base.FOOTER)

    def bye_message(self, bin):
        '''return a bye message'''
        msnlp = self.bye_message_msnlp()
        self.increase_base_id()

        header = BinHeader()
        header.identifier = self.base_id
        header.acknowledged_identifier = random_number(50000)
        header.total_data_size = len(msnlp) - 4
        header.message_length = len(msnlp) - 4
        header.flag = Base.BYE_FLAG
        
        self.last_received_header = bin
        self.last_sent_header = header
        self.status = Base.STATUS_BYE
        return header.data + msnlp
    
    def is_bye(self, bin, params, message):
        '''return true if the message is a bye from this p2p session'''
        return (contains(message, 'BYE MSNMSGR:') and \
            'To' in params and params['To'] == self.to_ and \
            'From' in params and params['From'] == self.from_ and \
            bin.identifier - 1 == self.his_base_id)

    # -------------------------- BYE ACK -----------------------------

    def bye_ack(self, bin):
        '''return a bye ack message'''
        self.increase_base_id()
        
        header = BinHeader()
        header.identifier = self.base_id
        header.total_data_size = bin.total_data_size
        header.acknowledged_data_size = bin.total_data_size
        header.flag = Base.BYE_ACK_FLAG
        header.acknowledged_identifier = bin.identifier
        header.acknowledged_unique_id = bin.acknowledged_identifier
        
        self.last_received_header = bin
        self.last_sent_header = header
        self.status = Base.STATUS_BYE_ACK
        return header.data + Base.FOOTER
        
    def is_bye_ack(self, bin):
        '''return True if bin is a bye ack bin header of
        this p2p session'''
        lsh = self.last_sent_header
        #lsh.total_data_size == bin.total_data_size and \

        if bin.flag == Base.BYE_ACK_FLAG and \
           lsh.total_data_size == bin.acknowledged_data_size and \
           lsh.identifier == bin.acknowledged_identifier and \
           lsh.acknowledged_identifier == bin.acknowledged_unique_id:
            return True

        return False

class Sender(Base):
    '''a class that send p2p content'''

    def __init__(self, p2p, bin_header, parameters, data):
        '''the switchboard that received the message,
        a BinHeader instance with the bin header content
        and a dict with the msnlp parameters
        data is the data to send'''
        Base.__init__(self)
        self.temporary_session = True

        self.last_received_header = bin_header
        self.his_base_id = bin_header.identifier

        if 'To' in parameters:
            self.to_ = get_to(parameters['To'])

        if 'From' in parameters:
            self.from_ = get_from(parameters['From'])

        self.data = data

        if 'Via' in parameters:
            self.via = get_via_branch(parameters['Via'])

        if 'Call-ID' in parameters:
            self.call_id = get_call_id(parameters['Call-ID'])

        if 'SessionID' in parameters:
            self.his_session_id = parameters['SessionID']

        if 'Conn-Type' in parameters:
            common.debug('direct connect, refusing', 'p2p')
            self.emit('msnp2p-message-ready', 
                self.decline_message(), self.from_)
            
        # we see if some value is an empty string
        if not self.to_ or \
            not self.from_ or \
            not self.via or \
            not self.call_id or \
            not self.session_id:
            common.debug('invalid msnlp header creating Sender', 'p2p')
            self.status = Base.STATUS_ERROR
        else:
            self.msg_received_id = p2p.connect('msnp2p-message-received', 
                self.on_msnp2p_message_received)
            self.msg_ready_id = self.connect('msnp2p-message-ready',
                p2p.on_msnp2p_message_ready)

            self.emit('msnp2p-message-ready',
                self.base_id_message(bin_header), self.from_)
            self.emit('msnp2p-message-ready', self.ok_message(), self.from_)

    def on_msnp2p_message_received(self, p2p, bin, parameters, message):
        '''method called when the switchboard receive a msnp2p message
        we check if it is from this session, if it is, we process it
        if we are finished or something went wrong, we disconnect the signal
        '''
        self.check_time_limit()
        if self.status == Base.STATUS_ERROR or \
            self.status == Base.STATUS_END:
            common.debug('disconnecting p2p handler', 'p2p')
            p2p.disconnect(self.msg_received_id)
            self.disconnect(self.msg_ready_id)
        else:
            common.debug('Sender checking', 'p2p')
            if self.status == Base.STATUS_200OK and \
                self.is_ok_ack(bin):
                common.debug('is 200 ok ack', 'p2p')
                self.emit('msnp2p-message-ready', 
                            self.dpm_message(bin), self.from_)
            elif self.status == Base.STATUS_PREPARATION and \
                    self.is_dpm_ack(bin):
                common.debug('is data preparation message ack', 'p2p')
                common.debug('sending data', 'p2p')
                for data_msg in self.data_message():
                    self.emit('msnp2p-message-ready', data_msg, self.from_)
            elif self.status == Base.STATUS_SENDING and \
                    self.is_data_ack(bin):
                common.debug('is data ack', 'p2p')
                self.emit('msnp2p-message-ready', 
                            self.bye_message(bin), self.from_)
                # we should wait the bye ack, but it make no sense
                # because there is nothing else to do
                self.status = Base.STATUS_END

class Receiver(Base):
    '''a class that request data to other client and receive it'''

    def __init__(self, p2p, from_, msnobj):
        '''from_ is our email, to is extracted from the msnobj, 
        msnobj is a Msnobj instance'''
        Base.__init__(self)
        self.temporary_session = True

        self.msnobj = msnobj

        self.from_ = from_
        self.to_ = msnobj.creator
        self.data = Buffer(msnobj.filename,'')

        if self.msnobj.type == Msnobj.Msnobj.DISPLAY_PICTURE:
            self.app_id = '1' 
        elif self.msnobj.type == Msnobj.Msnobj.CUSTOM_EMOTICON:
            self.app_id = '12' 
            
        self.msg_received_id = p2p.connect('msnp2p-message-received',
            self.on_msnp2p_message_received)
        self.msg_ready_id = self.connect('msnp2p-message-ready',
            p2p.on_msnp2p_message_ready)

        context = base64.b64encode(str(msnobj) + '\0')
        message = self.invite_message(context)
        self.emit('msnp2p-message-ready', message, self.to_)
    
    def on_msnp2p_message_received(self, p2p, bin, parameters, message):
        '''method called when the switchboard receive a msnp2p message
        we check if it is from this session, if it is, we process it
        if we are finished or something went wrong, we disconnect the signal
        '''
        self.check_time_limit()
        if self.status == Base.STATUS_ERROR or \
            self.status == Base.STATUS_END:
            common.debug('disconnecting p2p handler', 'p2p')
            p2p.disconnect(self.msg_received_id)
            self.disconnect(self.msg_ready_id)
        else:
            common.debug('receiver checking', 'p2p')
            if self.is_base_id_message(bin):
                common.debug('base id message', 'p2p')
                self.his_base_id = bin.identifier
            elif self.is_ok(bin, parameters, message):
                common.debug('200 ok message', 'p2p')
                self.emit('msnp2p-message-ready', 
                            self.ok_ack_message(bin), self.to_)
            elif self.is_dpm(bin):
                common.debug('data preparation message', 'p2p')
                self.emit('msnp2p-message-ready', 
                            self.dpm_ack_message(bin), self.to_)
            elif self.is_data(bin):
                common.debug('data message', 'p2p')
                # aMSN 0.97b loves to send the data twice
                if self.data.pointer != bin.data_offset:
                    common.debug('self.data.pointer != bin.data_offset')
                    return

                data = get_data_chunk(message, bin)
                # nice line :P
                self.data.data += data
                self.data.pointer += bin.message_length

                if self.is_last_data(bin):
                    common.debug('is last data', 'p2p')
                    self.emit('msnp2p-message-ready', 
                                self.data_ack(bin), self.to_)

                    if self.data.pointer != bin.total_data_size:
                        common.debug('corrupt data, not all pieces received', 'p2p')
                        print self.data.pointer, bin.total_data_size
                    elif self.msnobj.type == Msnobj.Msnobj.CUSTOM_EMOTICON:
                        common.debug('custom-emoticon-data-received', 'p2p')
                        p2p.emit('custom-emoticon-data-received', 
                            self.msnobj, self.data.data, self.to_)
                    elif self.msnobj.type == Msnobj.Msnobj.DISPLAY_PICTURE:
                        common.debug('display-picture-received', 'p2p')
                        p2p.emit('display-picture-received', 
                            self.msnobj, self.data.data, self.to_)
                    else:
                        common.debug('no CE or DP?', 'p2p')
                        print self.msnobj.type

            elif self.is_bye(bin, parameters, message):
                common.debug('bye message', 'p2p')
                self.emit('msnp2p-message-ready', self.bye_ack(bin))
                self.status = Base.STATUS_END

class FTReceiver(Base):
    '''a class to accept and handle an incoming file transfer'''

    def __init__(self, p2p, context, parameters, bin_header, data):
        '''context, parameters and bin_header are from the INVITE message,
        data is a Buffer instance'''
        Base.__init__(self)
        self.temporary_session = False

        self.context = context

        self.last_received_header = bin_header

        self.identifier = bin_header.identifier
        self.base_id = self.identifier
        self.his_base_id = self.identifier

        if 'To' in parameters:
            self.to_ = get_to(parameters['To'])

        if 'From' in parameters:
            self.from_ = get_from(parameters['From'])

        self.data = data

        if 'Via' in parameters:
            self.via = get_via_branch(parameters['Via'])

        if 'Call-ID' in parameters:
            self.call_id = get_call_id(parameters['Call-ID'])

        if 'SessionID' in parameters:
            self.session_id = int(parameters['SessionID'])

        # we see if some value is an empty string
        if not self.to_ or \
            not self.from_ or \
            not self.via or \
            not self.call_id or \
            not self.session_id:
            common.debug('invalid msnlp header creating FTReceiver', 'p2p')
            self.status = Base.STATUS_ERROR
        else:
            self.msg_received_id = p2p.connect('msnp2p-message-received', 
                self.on_msnp2p_message_received)
            self.msg_ready_id = self.connect('msnp2p-message-ready',
                p2p.on_msnp2p_message_ready)
        
        common.debug('sending FT ack', 'p2p')
        message = self.standard_ack(bin_header).data + Base.FOOTER
        self.last_sent_header = get_bin_header(message)
        self.emit('msnp2p-message-ready', message, self.from_)

        self.accept_id = p2p.connect('file-transfer-accepted', self.accept_ft)
        
        p2p.manager.msn.emit( 'new-conversation', \
            p2p.switchboard.firstUser, p2p.switchboard )

        p2p.emit('file-transfer-invite', int(self.session_id), \
            self.context, self.from_)
        
    def accept_ft(self, p2p, session, context, sender):
        '''callback for switchboards file-transfer-accepted, sends a 200ok
        message starting the transfer'''
        if int(session) == int(self.session_id) and \
           context == self.context and \
           sender == self.from_:

            #accept the file
            common.debug('accepting FT', 'p2p')
            message = self.ok_message()
            self.last_sent_header = get_bin_header(message)
            self.emit('msnp2p-message-ready', message, self.from_)
            p2p.disconnect(self.accept_id)
            self.accept_id = 0
        else:
            common.debug('its not my FT', 'p2p')

    def increase_base_id(self):
        '''this function is simpler with FT, so we override it'''
        self.base_id += 1

    def on_msnp2p_message_received(self, p2p, bin, parameters, message):
        '''method called when the switchboard receive a msnp2p message
        we check if it is from this session, if it is, we process it
        if we are finished or something went wrong, we disconnect the signal
        '''
        
        self.check_time_limit()
        
        if self.status == Base.STATUS_ERROR or \
            self.status == Base.STATUS_END:
            self.disconnect_handler(p2p)
        else:
            common.debug('FT Receiver checking', 'p2p')
            if 'SessionID' in parameters:
                session_id = parameters['SessionID']
            else:
                session_id = bin.session_id

            # update his base id and tell if it is the next identifier
            if bin.identifier == self.his_base_id + 1:
                # next message
                self.his_base_id = bin.identifier
                is_next = True
            elif bin.identifier <= self.his_base_id and \
                 bin.identifier > self.identifier:
                # old multipart mesage
                is_next = False
            else:
                is_next = False

            if int(session_id) == int(self.session_id) and \
               self.is_data(bin) and bin.flag == Base.FILE_FLAG:
                common.debug('data message', 'p2p')
                # aMSN 0.97b loves to send the data twice
                # TODO: check if this applies also to FT
                if self.data.pointer != bin.data_offset:
                    common.debug('self.data.pointer != bin.data_offset')
                    return
                
                data = get_data_chunk(message, bin)
                # nice line :P
                self.data.data += data
                self.data.pointer += bin.message_length

                self.status = Base.STATUS_RECEIVING
                p2p.emit('file-transfer-progress', int(self.session_id),
                    self.context, int(self.data.pointer), self.from_)
                
                if self.is_last_data(bin):
                    common.debug('is last data -- sending ack', 'p2p')
                    
                    # send data ack (TODO: compare)
                    self.emit('msnp2p-message-ready', \
                        self.data_ack(bin), self.from_)

                    if self.data.pointer != bin.total_data_size:
                        # transfer failed
                        common.debug('corrupt data, not all pieces received', \
                            'p2p')
                        common.debug(str(self.data.pointer) + ' ' + \
                            str(bin.total_data_size))
                        
                        p2p.emit('file-transfer-failed', int(self.session_id),\
                            self.context, 'corrupt', self.from_)
                        self.status = Base.STATUS_ERROR
                        self.disconnect_handler(p2p)
                    else:
                        # file received successfully
                        common.debug('file-transfer-received', 'p2p')
                        p2p.emit('file-transfer-received',
                            int(self.session_id), self.context,
                            self.data, self.from_)
                        # close session if no bye received
                        self.time_stamp = time.time()
                        self.temporary_session = True

            elif self.is_ack(bin):
                common.debug('is FTReceiver ack - ignored, useless', 'p2p')
                
            elif int(session_id) == int(self.session_id) and \
                 contains(message, 'MSNSLP/1.0 603 Decline'):

                common.debug("decline, cancel before accept", 'p2p')
                self.status = Base.STATUS_END
                # TODO: should we reply?
                p2p.emit('file-transfer-failed', int(self.session_id), 
                    self.context, 'cancelled', self.from_)
                self.disconnect_handler(p2p)

            elif is_next and contains(message, 'BYE MSNMSGR:'):
                common.debug('bye message', 'p2p')
                self.emit('msnp2p-message-ready', self.bye_ack(bin), self.from_)
                self.status = Base.STATUS_END
                if not self.temporary_session:
                    # we weren't waiting for a bye message
                    common.debug('transfer not finished, file cancelled', 'p2p')
                    p2p.emit('file-transfer-failed', int(self.session_id), 
                        self.context, 'cancelled', self.from_)
                    self.disconnect_handler(p2p)
            else:
                if int(session_id) == int(self.session_id):
                    common.debug('XXX unrecognized message')
                    if bin.flag == Base.ACK_FLAG:
                        common.debug("...but it's an ack, so you shouldn't worry")

    def disconnect_handler(self, p2p):
        common.debug('disconnecting p2p handler', 'p2p')
        self.disconnect(self.msg_ready_id)
        p2p.disconnect(self.msg_received_id)
        if self.accept_id:
            p2p.disconnect(self.accept_id)

class Buffer(object):
    '''a class that contain data to be readed or written'''
    
    def __init__(self, path, data):
        '''data is a place to put data or tu read data, path can
        be the path where we are going to write, or the path where
        the data comes from'''

        self.data = data
        self.path = path
        self.pointer = 0

    def size(self):
        '''return the size of the buffer'''
        return len(self.data)

    def read(self):
        '''read from path and put it on self.data'''
        try:
            file_descriptor = open(self.path, 'rb')
            self.data = file_descriptor.read()
            file_descriptor.close()
        except IOError, ex:
            print 'cant read from %s' % self.path
            print str(ex)
    
    def write(self):
        '''write self.data to self.path'''
        try:
            file_descriptor = open(self.path, 'wb')
            file_descriptor.write(self.data)
            file_descriptor.close()
        except IOError, ex:
            print 'cant write to %s' % self.path
            print str(ex)

    def get_chunks(self, chunk_size):
        '''return a list of n pieces of size, the last can be smaller'''

        if self.size() < chunk_size:
            return [self.data]

        chunk_list = []
        # start at 0, max self.size(), step chunk_size
        for pointer in range(0, self.size(), chunk_size):
            if (pointer + chunk_size) > self.size():
                chunk_list.append(self.data[pointer:self.size()])
            else:
                chunk_list.append(self.data[pointer:pointer + chunk_size])

        return chunk_list


class ByteArray(object):
    '''a class that represent a byte array (little endian)'''
    
    def __init__(self, size):
        '''create a byte array with size=size and initialize with 0'''
        
        self.fmt = '<' + str(size) + 'B'
        self.data = struct.pack(self.fmt, *[0 for i in range(size)])
        self.value_list = list(struct.unpack(self.fmt, self.data))

    def __setitem__(self, pos, value):
        '''set the value in pos'''
        
        if len(self.value_list) > pos:
            self.value_list[pos] = value
            self.data = struct.pack(self.fmt, *self.value_list)
            
    def __getitem__(self, pos):
        '''return the byte at pos in the byte array'''
        return self.value_list[pos]
        
    def set_dword(self, pos, value):
        '''set a dword at post with value'''
        # less significative
        byte1 = self.__extract_byte(value, 0)
        byte2 = self.__extract_byte(value, 1)
        byte3 = self.__extract_byte(value, 2)
        byte4 = self.__extract_byte(value, 3)
        
        self[pos] = byte1
        self[pos + 1] = byte2
        self[pos + 2] = byte3
        self[pos + 3] = byte4
        
    def get_dword(self, pos):
        '''return the dword at pos'''
        byte1 = self[pos]
        byte2 = self[pos + 1]
        byte3 = self[pos + 2]
        byte4 = self[pos + 3]
        
        return (byte4<<24) + (byte3<<16) + (byte2<<8) + byte1
        
    def set_qword(self, pos, value):
        '''set the qword at post with value'''
        byte1 = self.__extract_byte(value, 0)
        byte2 = self.__extract_byte(value, 1)
        byte3 = self.__extract_byte(value, 2)
        byte4 = self.__extract_byte(value, 3)
        
        byte5 = self.__extract_byte(value, 4)
        byte6 = self.__extract_byte(value, 5)
        byte7 = self.__extract_byte(value, 6)
        byte8 = self.__extract_byte(value, 7)
        
        self[pos] = byte1
        self[pos + 1] = byte2
        self[pos + 2] = byte3
        self[pos + 3] = byte4
        
        self[pos + 4] = byte5
        self[pos + 5] = byte6
        self[pos + 6] = byte7
        self[pos + 7] = byte8
        
    def get_qword(self, start):
        '''return the qword at pos'''
        return (self.get_dword(start + 4) << 32) + self.get_dword(start)
    
    def __extract_byte(self , number , byte):
        '''extract a given byte of a number and return it
        0 -> first byte ( less significative )
        '''
        
        return (number & (255 << byte * 8)) >> (byte * 8)
        
    def __repr__(self):
        '''return the representation of this object'''
        return self.data
    
    def __len__(self):
        '''return the length of this object'''
        return len(self.value_list)

            
class BinHeader(ByteArray):
    '''this class represent the bin header'''
    
    def __init__(self, data=None):
        '''Contructor'''
        ByteArray.__init__(self, 48)
        
        if data != None:
            self.fill(data)
            
    def set_session_id(self, dword):
        '''set the value of session id'''
        self.set_dword(0, dword)
        
    def get_session_id(self):
        '''get the value of session id'''
        return self.get_dword(0)
    
    session_id = property(fset=set_session_id, fget=get_session_id)

    def set_identifier(self, dword):
        '''set the value of identifier'''
        self.set_dword(4, dword)
        
    def get_identifier(self):
        '''get the value of identifier'''
        return self.get_dword(4)
   
    identifier = property(fset=set_identifier, fget=get_identifier)

    def set_data_offset(self, qword):
        '''set the value of data offset'''
        self.set_qword(8, qword)
        
    def get_data_offset(self):
        '''get the value of data offset'''
        return self.get_qword(8)
    
    data_offset = property(fset=set_data_offset, fget=get_data_offset)

    def set_total_data_size(self, qword):
        '''set the value of total data size'''
        self.set_qword(16, qword)
        
    def get_total_data_size(self):
        '''get the value of total data size'''
        return self.get_qword(16)
    
    total_data_size = property(fset=set_total_data_size, 
                                fget=get_total_data_size)

    def set_message_length(self, dword):
        '''set the value of message length'''
        self.set_dword(24, dword)
        
    def get_message_length(self):
        '''get the value of message length'''
        return self.get_dword(24)

    message_length = property(fset=set_message_length, 
                                fget=get_message_length)
    
    def set_flag(self, dword):
        '''set the value of flag'''
        self.set_dword(28, dword)
        
    def get_flag(self):
        '''get the value of flag'''
        return self.get_dword(28)
    
    flag = property(fset=set_flag, fget=get_flag)

    def set_acknowledged_identifier(self, dword):
        '''set the value of acknowledged identifier'''
        self.set_dword(32, dword)
        
    def get_acknowledged_identifier(self):
        '''get the value of identifier'''
        return self.get_dword(32 )

    acknowledged_identifier = property(fset=set_acknowledged_identifier, 
                                        fget=get_acknowledged_identifier)
    
    def set_acknowledged_unique_id(self, dword):
        '''set the value of acknowledged unique id'''
        self.set_dword(36, dword)
        
    def get_acknowledged_unique_id(self):
        '''get the value of acknowledged unique id'''
        return self.get_dword(36)

    acknowledged_unique_id = property(fset=set_acknowledged_unique_id, 
                                        fget=get_acknowledged_unique_id)
    
    def set_acknowledged_data_size(self, qword):
        '''set the value of acknowledged data size'''
        self.set_qword(40, qword)
        
    def get_acknowledged_data_size(self):
        '''get the value of acknowledged data size'''
        return self.get_qword(40)

    acknowledged_data_size = property(fset=set_acknowledged_data_size, 
                                        fget=get_acknowledged_data_size)    
    
    def print_numbers(self):
        '''print a nice string with the values of the array'''
        count = 0
        for i in self.data:
            print hex(ord(i)),
            
            # :)
            if len(str(hex(ord(i)))) == 3:
                print '  ',
            else:
                print ' ',
                
            count += 1
            if count > 15:
                print
                count = 0
                
        print
        
    def print_fields(self):
        '''print the binary fields'''
        print '-------------------------------------------\n'
        print 'SessionID:               ' + str(self.session_id)
        print 'Identifier:              ' + str(self.identifier)
        print 'Data offset:             ' + str(self.data_offset)
        print 'Total data size:         ' + str(self.total_data_size)
        print 'Message length:          ' + str(self.message_length)
        print 'Flag:                    ' + str(self.flag)
        print 'Acknowledged identifier: ' + str(self.acknowledged_identifier)
        print 'Acknowledged unique ID:  ' + str(self.acknowledged_unique_id)
        print 'Acknowledged data size:  ' + str(self.acknowledged_data_size)
        print '-------------------------------------------\n'
        
    def fill(self, data):
        '''fill the object with the data provided'''
        self.set_session_id(get_dword_from_data(data, 0))
        self.set_identifier(get_dword_from_data(data, 4))
        self.set_data_offset(get_qword_from_data(data, 8))
        self.set_total_data_size(get_qword_from_data(data, 16))
        self.set_message_length(get_dword_from_data(data, 24))
        self.set_flag(get_dword_from_data(data, 28))
        self.set_acknowledged_identifier(get_dword_from_data(data, 32))
        self.set_acknowledged_unique_id(get_dword_from_data(data, 36))
        self.set_acknowledged_data_size(get_qword_from_data(data, 40))

class FTContext(ByteArray):
    '''this class represent File Transfer Context'''
    
    def __init__(self, data=None):
        '''Constructor'''
        ByteArray.__init__(self, 25)
        self.__filename = ''
        self.name = ''
        self.crap = '\xff\xff'
        self.set_unknown_dword( 0xffff0000 )
        self.set_header_length( 574 )
        self.set_preview('')
        if data != None:
            self.fill(data)
            
            
    def __repr__(self):
        '''return the representation of this object'''
        #cxt = self.data[0:18]+self.__filename+self.data[19:]+self.preview
        cxt = self.data[0:20]+self.__filename+self.data[21:25]+self.crap+self.preview
        
        return cxt
            
    def set_header_length(self, dword):
        '''set the value of session id'''
        self.set_dword(0, dword)
        
    def get_header_length(self):
        '''get the value of session id'''
        return self.get_dword(0)
    
    header_length = property(fset=set_header_length, fget=get_header_length)

    def set_msnc(self, dword):
        '''set the value of identifier'''
        self.set_dword(4, dword)
        
    def get_msnc(self):
        '''get the value of identifier'''
        return self.get_dword(4)
   
    msnc = property(fset=set_msnc, fget=get_msnc)

    def set_file_size(self, qword):
        '''set the value of data offset'''
        self.set_dword(8, qword)
        
    def get_file_size(self):
        '''get the value of data offset'''
        return self.get_dword(8)
    
    file_size = property(fset=set_file_size, fget=get_file_size)

    def set_data_type(self, dword):
        '''set the value of total data size'''
        self.set_dword(15, dword)
        
    def get_data_type(self):
        '''get the value of total data size'''
        return self.get_dword(15)
    
    data_type = property(fset=set_data_type, fget=get_data_type)

    def set_filename(self, data):
        '''set the value of filename'''
        data = data+'\0'*(548-len(data))
        self.__filename = data.encode('utf-16-le')
        
    def get_filename(self):
        '''get the value of filename'''
        return self.__filename.decode('utf-16-le').replace('\0', '')

    filename = property(fset=set_filename, fget=get_filename)
    
    def set_unknown_dword(self, dword):
        '''set the value of total data size'''
        self.set_dword(21, dword)
        
    def get_unknown_dword(self):
        '''get the value of total data size'''
        return self.get_dword(21)
        
    unknown_dword = property(fset=set_unknown_dword, fget=get_unknown_dword)
    
    def set_preview(self, data):
        '''set the value of preview'''
        self.__preview = data
        
    def get_preview(self):
        '''get the value of preview'''
        return self.__preview

    preview = property(fset=set_preview, fget=get_preview)
    
    def print_numbers(self):
        '''print a nice string with the values of the array'''
        count = 0
        for i in self.data:
            print hex(ord(i)),
            
            # :)
            if len(str(hex(ord(i)))) == 3:
                print '  ',
            else:
                print ' ',
                
            count += 1
            if count > 15:
                print
                count = 0
                
        print
        
    def print_fields(self):
        '''print the binary fields'''
        print '-------------------------------------------\n'
        print 'Header length:           ' + str(self.header_length)
        print 'msnc:                    ' + str(self.msnc)
        print 'File size:               ' + str(self.file_size)
        print 'Type:                    ' + str(self.data_type)
        print 'Unknown Dword:            ' + str(self.unknown_dword)
        print 'Filename:                ' + str(self.filename)
        #print 'Preview:                 ' + str(self.preview).__repr__()
        print '-------------------------------------------\n'
        #print self.__filename.__repr__()
        
    def fill(self, data):
        '''fill the object with the data provided'''
        self.set_header_length( get_dword_from_data(data, 0) )
        self.set_msnc( get_dword_from_data(data, 4) )
        self.set_file_size( get_dword_from_data(data, 8) )
        self.set_data_type( get_dword_from_data(data, 16) )
        self.__filename = data[20:540]
        self.set_unknown_dword( get_dword_from_data(data, 569) )
        self.crap = data[573:self.header_length]
        #print data[568:572].__repr__()
        #print data[572:574].__repr__()
        self.preview = data[self.header_length:]

gobject.type_register(Base)
gobject.type_register(Sender)
gobject.type_register(Receiver)
