# -*- coding: utf-8 -*-

'''Logger module, contains the Logger class and a test case'''

#   This file is part of emesene.
#
#    Emesene is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    Emesene is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with emesene; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301
#    USA

ERROR = False

import os
import time
import urllib
import gobject

try:
    from pysqlite2 import dbapi2 as sqlite
except ImportError:
    ERROR = True

import Plugin

class Logger(object):
    '''this class represent a event logger for conversations
    its made for emesene but can be used on any other im application
    or similar'''

    CREATE_USER = '''
    CREATE TABLE user
    (
        id INTEGER PRIMARY KEY,
        account TEXT
 )
    '''

    CREATE_CONVERSATION = '''
    CREATE TABLE conversation
    (
        id INTEGER PRIMARY KEY,
        started INTEGER
 )
    '''

    CREATE_EVENT = '''
    CREATE TABLE event
    (
        id INTEGER PRIMARY KEY,
        stamp INTEGER,
        name TEXT
 )
    '''

    CREATE_USER_EVENT = '''
    CREATE TABLE user_event
    (
        id_event INTEGER,
        id_user INTEGER,
        data TEXT
 )
    '''

    CREATE_CONVERSATION_EVENT = '''
    CREATE TABLE conversation_event
    (
        id_event INTEGER,
        id_conversation INTEGER,
        id_user INTEGER,
        data TEXT
 )
    '''

    USER_EXISTS = 'SELECT id FROM user WHERE account = ?'
    USER_ADD = 'INSERT INTO user(id, account) VALUES(null, ?)'

    USER_EVENT_ADD = '''INSERT INTO user_event(id_event, id_user, data)
        VALUES(?, ?, ?)'''

    EVENT_ADD = 'INSERT INTO event(id, stamp, name) VALUES(null, ?, ?)'

    CONVERSATION_EVENT_ADD = '''INSERT INTO conversation_event(
        id_event, id_conversation, id_user, data) VALUES(?, ?, ?, ?)'''

    CONVERSATION_EXISTS = 'SELECT id FROM conversation WHERE started = ?'
    CONVERSATION_ADD = 'INSERT INTO conversation(id, started) \
                        VALUES(null, ?)'

    def __init__(self, path):
        '''constructor, try to open the database at path, if can not open
        it, then try to create it'''
    
        self.path = path
        self.connection = sqlite.connect(path)
        self.cursor = self.connection.cursor()

        try:
            self._create()
        except sqlite.OperationalError:
            # pylint, are you happy? :P
            print 'database already there'

    def _create(self):
        '''create the tables'''

        self.cursor.execute(Logger.CREATE_USER)
        self.cursor.execute(Logger.CREATE_CONVERSATION)
        self.cursor.execute(Logger.CREATE_EVENT)
        self.cursor.execute(Logger.CREATE_USER_EVENT)
        self.cursor.execute(Logger.CREATE_CONVERSATION_EVENT)
        self.connection.commit()

    def user_exists(self, account):
        '''return the user id if the account exists -1 otherwise'''

        self.cursor.execute(Logger.USER_EXISTS, (account,))

        result = self.cursor.fetchall() 
        if len(result) >= 1:
            return result[0][0]
        else:
            return -1

    def user_add(self, account):
        '''try to add an user if account exists don't add it,
        return the id in both cases'''

        user_id = self.user_exists(account)

        if user_id == -1:
            self.cursor.execute(Logger.USER_ADD, (account,))
            self.connection.commit()
            return self.cursor.lastrowid

        return user_id

    def user_event_add(self, name, account, data):
        '''add a user event with the event name "name", generated by
        "account" and containing "data", return the id'''

        user_id = self.user_add(account)
        event_id = self.event_add(name)
        self.cursor.execute(Logger.USER_EVENT_ADD, (event_id, 
                             user_id, data))
        self.connection.commit()
        return self.cursor.lastrowid

    def event_add(self, name):
        '''add an event with name "name", return the id'''

        self.cursor.execute(Logger.EVENT_ADD, (time.time(), name))
        self.connection.commit()
        return self.cursor.lastrowid

    def conversation_event_add(self, name, started, account, data):
        '''add a conversation event with name "name" started at the
        "started" time, with data "data", return the id'''

        conversation_id = self.conversation_add(started)
        event_id = self.event_add(name)
        user_id = self.user_add(account)
        self.cursor.execute(Logger.CONVERSATION_EVENT_ADD, (event_id,
                                conversation_id, user_id, data))
        self.connection.commit()
        return self.cursor.lastrowid

    def conversation_exists(self, started):
        '''return the id of the conversation if exists, -1 otherwise'''

        self.cursor.execute(Logger.CONVERSATION_EXISTS, (started,))

        result = self.cursor.fetchall() 
        if len(result) >= 1:
            return result[0][0]
        else:
            return -1

    def conversation_add(self, started):
        '''try to add a conversation if exists don't add it,
        return the id in both cases'''

        conversation_id = self.conversation_exists(started)

        if conversation_id == -1:
            self.cursor.execute(Logger.CONVERSATION_ADD, (started,))
            self.connection.commit()
            return self.cursor.lastrowid

        return conversation_id

    def query(self, query):
        '''make a custom quera, return a tuple with all the results
        the exception are not handled, because you need to know if a
        query went bad on your plugin not this one, so please use 
        try except blocks'''
        
        self.cursor.execute(query)
        return self.cursor.fetchall()

class MainClass(Plugin.Plugin):
    '''Main plugin class'''
    
    def __init__(self, controller, msn):
        '''Contructor'''
        Plugin.Plugin.__init__(self, controller, msn, 1000)
        
        self.description = 'Log all events to a file'
        self.authors = {'Mariano Guerra' : 
            'luismarianoguerra at gmail dot com'}
        self.website = 'http://emesene.org'
        self.displayName =  'Logger'
        self.name = 'Logger'
        
        self.config = controller.getConfig()
        self.config.readPluginConfig(self.name)
        self.path = self.config.getPluginValue(self.name, 
            'path', self.msn.cacheDir)

        self.file_name = os.path.join(self.path, self.msn.user + '.db')
        self.logger = Logger(self.file_name)
        self.ids = []

        self.signals = {}
        self.signals['self-nick-changed'] = self._cb_self_nick_changed
        self.signals['personal-message-changed'] = \
            self._cb_personal_message_changed
        self.signals['self-status-changed'] = self._cb_self_status_changed
        self.signals['self-personal-message-changed'] = \
            self._cb_self_personal_message_changed
        self.signals['self-current-media-changed'] = \
            self._cb_self_current_media_changed
        self.signals['nick-changed'] = self._cb_nick_changed
        self.signals['contact-status-change'] = \
            self._cb_contact_status_change
        self.signals['user-offline'] = self._cb_user_offline
        self.signals['new-switchboard'] = self._cb_new_switchboard
        self.signals['display-picture-changed'] = \
            self._cb_display_picture_changed

        self.sb_ids = []
        self.sb_signals = {}
        self.sb_signals['user-join'] = self._cb_sb_user_join
        self.sb_signals['user-leave'] = self._cb_sb_user_leave
        self.sb_signals['user-offline'] = self._cb_sb_user_offline
        self.sb_signals['user-online'] = self._cb_sb_user_online
        self.sb_signals['message'] = self._cb_sb_message
        self.sb_signals['action-message'] = self._cb_sb_action_message
        self.sb_signals['ink-message'] = self._cb_sb_ink_message
        self.sb_signals['nudge'] = self._cb_sb_nudge
        self.sb_signals['message-sent'] = self._cb_sb_message_sent
        self.sb_signals['action-sent'] = self._cb_sb_action_message_sent
        self.sb_signals['ink-sent'] = self._cb_sb_ink_message_sent
        self.sb_signals['nudge-sent'] = self._cb_sb_nudge_sent
        self.sb_signals['custom-emoticon-received'] = \
            self._cb_sb_custom_emoticon_received

        # timeout source to disconect it
        self.to_source = None
        all_events = ','.join(self.signals.keys() + self.sb_signals.keys())
        
        self.events_enabled =  self.config.getPluginValue(self.name, 
            'eventsEnabled', all_events).replace(' ', '').split(',')
    
    def start(self):
        '''start the plugin'''

        self.enabled = True
        
        for (key, value) in self.signals.iteritems():
            if key in self.events_enabled:
                self.ids.append(self.msn.connect(key, value))

        self.to_source = gobject.timeout_add(self.period, self.action)
        
    def stop(self):    
        '''stop the plugin'''
       
        # disconnect msn signals
        for identifier in self.ids:
            self.msn.disconnect(identifier)
        
        # disconnect switchboard signals
        for (switchboard, identifiers) in self.sb_ids:
            for identifier in identifiers:
                switchboard.disconnect(identifier)

        self.enabled = False

        gobject.source_remove(self.to_source)
        
    def action(self):
        '''clean the switchboard list from closed switchboards'''

        new_list = []

        for (switchboard, identifiers) in self.sb_ids:
            if switchboard.status == 'closed':
                for identifier in identifiers:
                    switchboard.disconnect(identifier)
            else:
                new_list.append((switchboard, identifiers,))

        self.sb_ids = new_list
    
    def check(self):
        '''check if everything is OK to start the plugin
        return a tuple whith a boolean and a message
        if OK -> (True , 'some message')
        else -> (False , 'error message')'''
       
        if ERROR:
            return (False, 'pysqlite2 not available, please install it.')
        else:
            return (True, 'Ok')

    def get_last_status(self, account, num = 1):
        '''return the last status of a contact, if num is > 1
        then return the last num status, the format of the list is
        a list of lists with the timestamp and the status'''

        query = '''
        select e.stamp, ue.data 
        from event e, user_event ue, user u
        where e.id = ue.id_event and
        u.id = ue.id_user and
        e.name = "status-changed" and
        u.account = "%s"
        order by e.stamp desc
        limit %s
        '''

        return self.logger.query(query % (account, num))
            
    def get_last_nick(self, account, num = 1):
        '''return the last nick of a contact, if num is > 1
        then return the last num nicks, the format of the list is
        a list of lists with the timestamp and the nick'''

        query = '''
        select e.stamp, ue.data 
        from event e, user_event ue, user u
        where e.id = ue.id_event and
        u.id = ue.id_user and
        e.name = "nick-changed" and
        u.account = "%s"
        order by e.stamp desc
        limit %s
        '''

        return self.logger.query(query % (account, num))
            
    def get_last_personal_message(self, account, num = 1):
        '''return the last pm of a contact, if num is > 1
        then return the last num pms, the format of the list is
        a list of lists with the timestamp and the pm'''

        query = '''
        select e.stamp, ue.data 
        from event e, user_event ue, user u
        where e.id = ue.id_event and
        u.id = ue.id_user and
        e.name = "personal-message-changed" and
        u.account = "%s"
        order by e.stamp desc
        limit %s
        '''

        return self.logger.query(query % (account, num))
    
    def get_last_message(self, account, num = 1):
        '''return the last message of a contact, if num is > 1
        then return the last num messages, the format of the list is
        a list of lists with the timestamp and the message'''

        query = '''
        select e.stamp, ce.data 
        from event e, conversation_event ce, user u
        where e.id = ce.id_event and
        u.id = ce.id_user and
        e.name = "message" and
        u.account = "%s"
        order by e.stamp desc
        limit %s
        '''

        return self.logger.query(query % (account, num))
            
    def get_last_display_picture(self, account, num = 1):
        '''return the last dp of a contact, if num is > 1
        then return the last num dps, the format of the list is
        a list of lists with the timestamp and the dp'''

        query = '''
        select e.stamp, ue.data 
        from event e, user_event ue, user u
        where e.id = ue.id_event and
        u.id = ue.id_user and
        e.name = "display-picture" and
        u.account = "%s"
        order by e.stamp desc
        limit %s
        '''

        return self.logger.query(query % (account, num))
    
    def get_last_custom_emoticon(self, account, num = 1):
        '''return the last ce of a contact, if num is > 1
        then return the last num ces, the format of the list is
        a list of lists with the timestamp and the ce'''

        query = '''
        select e.stamp, ce.data 
        from event e, conversation_event ce, user u
        where e.id = ce.id_event and
        u.id = ce.id_user and
        e.name = "custom-emoticon" and
        u.account = "%s"
        order by e.stamp desc
        limit %s
        '''

        return self.logger.query(query % (account, num))

    def get_last_conversation(self, account, num = 1):
        '''return the last message that was said on a conversation where
        account was present, the format of the list is a list od lists
        with the timestamp, email and message'''

        # in the inner select we select all the id of conversations
        # where account was present, then on the outer conversation
        # we select all the messages from that conversations
        query = '''
        select e.stamp, u.account, ce.data
        from event e, conversation_event ce, user u 
        where e.id = ce.id_event and 
        u.id = ce.id_user and 
        ce.id_conversation in 
        (select distinct c.id 
        from conversation c, event e, conversation_event ce, user u 
        where c.id = ce.id_conversation and 
        e.id = ce.id_event and 
        ce.id_user = u.id and 
        u.account = "%s") and 
        e.name = "message" 
        order by e.stamp desc
        limit %s
        '''

        return self.logger.query(query % (account, num))

    # callbacks

    def _cb_new_switchboard(self, msn, switchboard):
        '''called when a new switchboard is created'''

        id_list = []
        
        for (key, value) in self.sb_signals.iteritems():
            if key in self.events_enabled:
                id_list.append(switchboard.connect(key, value))

        self.sb_ids.append((switchboard, id_list,))
    
    def _cb_self_nick_changed(self, msn, old, new):
        '''called when we change our own nick'''
        self.logger.user_event_add('self-nick-changed', self.msn.user, new)

    def _cb_personal_message_changed(self, msn, email, personal_message):
        '''called when a user change his personal message'''
        if personal_message != '':
            self.logger.user_event_add('personal-message-changed', email, 
                                        personal_message)

    def _cb_self_status_changed(self, msn, status):
        '''called when we change the status'''
        self.logger.user_event_add('status-changed', self.msn.user, status)

    def _cb_self_personal_message_changed(self, msn, ourmail, 
                                            personal_message):
        '''called when we change our personal message'''
        self.logger.user_event_add('personal-message-changed', ourmail, 
                                    personal_message)

    def _cb_self_current_media_changed(self, msn, ourmail, 
                                        personal_message, dic):
        '''called when we change our current media'''
        self.logger.user_event_add('personal-message-changed', ourmail, 
                                    personal_message)

    def _cb_nick_changed(self, msn, email, nick):
        '''called when someone change his nick'''
        self.logger.user_event_add('nick-changed', email, nick)

    def _cb_contact_status_change(self, msn, email, status):
        '''called when someone change his status'''
        self.logger.user_event_add('status-changed', email, status)

    def _cb_user_offline(self, msn, email):
        '''called when someone goes offline'''
        self.logger.user_event_add('status-changed', email, 'FLN')

    def _cb_sb_user_join(self, switchboard, mail):
        '''called when an user join to the conversation'''
        self.logger.conversation_event_add('user-join',  
            switchboard.started, mail, '')

    def _cb_sb_user_leave(self, switchboard, mail):
        '''called when an user leaves the conversation'''
        self.logger.conversation_event_add('user-leave',  
            switchboard.started, mail, '')

    def _cb_sb_user_offline(self, switchboard, mail):
        '''called when an user goes offline'''
        self.logger.conversation_event_add('user-offline', 
            switchboard.started, mail, '')

    def _cb_sb_user_online(self, switchboard, mail):
        '''called when an user come online'''
        self.logger.conversation_event_add('user-online',  
            switchboard.started, mail, '')

    def _cb_sb_message(self, switchboard, mail, nick, body, format, charset):
        '''called when we receive a message'''
        message = '%s\r\n%s\r\n%s' % (format, charset, body)
        self.logger.conversation_event_add('message',  
            switchboard.started, mail, message)

    def _cb_sb_action_message(self, switchboard, mail, data):
        '''called when an action message is received'''
        self.logger.conversation_event_add('action-message', 
            switchboard.started, mail, data)

    def _cb_sb_ink_message(self, switchboard, mail, filename):
        '''called when an ink message is received'''
        self.logger.conversation_event_add('ink-message', 
            switchboard.started, mail, filename)

    def _cb_sb_nudge(self, switchboard, mail):
        '''called when a nudge is received'''
        self.logger.conversation_event_add('nudge', switchboard.started,
            mail, '')

    def _cb_sb_message_sent(self, switchboard, body, format, charset):
        '''called when we send a message'''
        format = format.split('X-MMS-IM-Format: ')[1]
        message = '%s\r\n%s\r\n%s' % (format, charset, body)
        self.logger.conversation_event_add('message',  
            switchboard.started, self.msn.user, message)

    def _cb_sb_action_message_sent(self, switchboard, data):
        '''called when an action message is received'''
        self.logger.conversation_event_add('action-message', 
            switchboard.started, self.msn.user, data)

    def _cb_sb_ink_message_sent(self, switchboard):
        '''called when an ink message is sent'''
        self.logger.conversation_event_add('ink-message',  
            switchboard.started, self.msn.user, '')

    def _cb_sb_nudge_sent(self, switchboard):
        '''called when a nudge is sent'''
        self.logger.conversation_event_add('nudge', 
            switchboard.started, self.msn.user, '')

    def _cb_sb_custom_emoticon_received(self, switchboard, shortcut, msnobj):
        '''called when a custom emoticon is received (du'h)'''
        filename = shortcut + '_' + msnobj.sha1d + '.tmp'
        filename = urllib.quote(filename).replace('/', '_')
        complete_filename = self.msn.cacheDir + os.sep + filename
        self.logger.conversation_event_add('custom-emoticon', 
            switchboard.started, self.msn.user, shortcut + ' ' + complete_filename)

    def _cb_display_picture_changed(self, msn, switchboard, msnobj, mail):
        '''called when a new display picture is received'''
        contact = self.msn.contactManager.getContact(mail)
        filename = contact.displayPicturePath
        # basicaly, retrieve the last display picture path of this user
        # it must be a better way but im tired :S
        q1 = '''
        select max(ev.stamp) 
        from event ev, user_event uev, user us 
        where uev.id_event = ev.id 
        and uev.id_user = us.id 
        and us.id = uev.id_user
        and us.account = "%s"'''

        r = self.logger.query(q1 % (mail, ))[0][0]

        ugly_query = '''
        select ue.data 
        from user_event ue, event e, user u 
        where e.name = "display-picture" 
        and u.id = ue.id_user 
        and ue.id_event = e.id 
        and u.account = "%s" 
        and e.stamp = %s'''

        if r is not None:
            result = self.logger.query(ugly_query % (mail, r))

            if len(result) > 0 and str(result[0][0]) == filename:
                return

        self.logger.user_event_add('display-picture', mail, filename)

